/*
 * Copyright (c) 2004-2005 Sergey Lyubka <valenok@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/*
 * Simple and portable HTTP server, http://shttpd.sf.net
 *
 * Compilation:
 *    SSL support:		-DWITH_SSL -lssl -lcrypto
 *    No CGI:			-DNO_CGI
 *    No Authorization:		-DNO_AUTH
 *    PUT,DELETE support:	-DWITH_PUT_AND_DELETE
 *    Override max request size	-DIO_MAX=xxxx
 *
 * Changelog:
 *   1.1	Multiple index files support
 *   1.2	Redirecting errors to a file.
 *		Fixed 'page contains no data' bug.
 *   1.3	Better error reporting (404, 500)
 *		Fixed CGI handling for win32, when CGI file is a native
 *		executable, not a script
 *   1.4	All replies are now set to HTTP/1.1 instead of HTTP/1.0
 *		For regular files, added headers: Content-Length, Date,
 *		Last-Modified, Etag.
 *		Fixed win32 CGI: passing environment to spawaned script.
 *   1.5	After start, change current dir to docroot
 *		Optimized environment build for CGI. Added few more
 *		env variables (REMOTE_ADDR, USER_AGENT, PATH_INFO, etc)
 *		Access log file opened once at start time, not on every request
 *		Added GUI for Windows
 *   1.6	Fixed CGI headers output (newline was missed)
 *		Thanks to Arno Schuring and Frash.
 *   1.7	Directory Listing support.
 *		'Kernel' rewritten. there is no global select() anymore.
 *		win32 GUI settings are stored in HKCU, thanks Ben Truscott
 *   1.8	Basic authentication is gone. Digest auth (MD5) is used instead.
 *   1.9	Support for user-defined mime types file
 *   1.10	Support for PUT and DELETE methods. Put check for the method
 *		(function setmethod()). Memory leak fixed in redirect().
 *   1.11	Rewrite. Not using vectors for parsing. Windows GUI code
 *		was thrown out.
 *   1.12	Added embedded interface
 *   1.13	Refactoring. Using standard FILENAME_MAX for paths.
 *		implemented EMBEDDED MT case.
 *   1.14	All HTTP headers are exported to CGI as HTTP_* variables.
 *		DOCUMENT_ROOT, SERVER_NAME are exported. CGIs to execute in
 *		their own directories. Thanks to Yair Lenga.
 *		-u option now takes login name, not numeric UID. if -u option
 *		is set, setgid() and setuid() are done. Thanks to Ronald Reed.
 *		Fixed authorization code to work with IE
 *		added shttpd_protect_url()
 *   1.15	Fixed PATH_TRANSLATED variable.
 *		implemented -I option (inetd mode), thanks to Toby Peterson
 *   1.16	Typo introduced when inetd support was added: shttpd_poll
 *		did not set sockaddr length before accept(), fixed.
 *		added shttpd_template()
 *		Also, -DEMBEDDED build fixed.
 *   1.17	ARM build did not like headers parisng sscanf() format string,
 *		slightly reimplemented. Nicer directory listing, file sizes
 *		in human-readable format, thanks to Rasmus Andersson.
 *   1.18	bug in mysocketpair(), fix suggested by David Weibel
 *   1.19	Security holes are fixed. Thanks to Adam Zeldis for his
 *		excellent code review and report.
 */

#define	VERSION		"1.19"		/* Version */
#define	EXPIRE_TIME	30		/* Expiration time, seconds */
#ifndef IO_MAX
#define	IO_MAX		32768		/* Max request size */
#endif /* IO_MAX */
#define	METHOD_MAX	15		/* Max HTTP method and protocol size */
#define	USER_MAX	64		/* Remote user name maxsize */
#define	AUTH_MAX	1024		/* Authorization line */
#define	NELEMS(ar)	(sizeof(ar) / sizeof(ar[0]))

#define	_STR(x)		#x
#define	_STR2(x)	_STR(x)

#ifdef _WIN32					/* Windows specific */
#pragma comment(lib,"ws2_32")
#include <winsock.h>
#include <windows.h>
#include <process.h>
#include <io.h>
#define	ERRNO			GetLastError()
typedef unsigned int		uint32_t;
typedef unsigned short		uint16_t;
typedef	DWORD			pthread_t;
#define	O_NONBLOCK		0
#define	waitpid(a,b,c)		0
#define	EWOULDBLOCK		WSAEWOULDBLOCK
#define	snprintf		_snprintf
#define	vsnprintf		_vsnprintf
#define	setuid(a)		0
#define	pthread_create(a,b,c,d)	CreateThread(0, 0, c, d, 0, a)
#define	pthread_exit(x)		ExitThread(x)
#define	pthread_detach(x)	0
#define	_POSIX_
#define	__func__		__FILE__ ", line "_STR2(__LINE__) ":"

/* POSIX dirent interface */
struct dirent {
	char	*d_name;
};
typedef struct DIR {
	long			handle;
	struct _finddata_t	info;
	struct dirent		result;
	char			*name;
} DIR;

#else						/* UNIX specific */

#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>		/* Some linuxes put struct timeval there */

#include <pwd.h>
#include <unistd.h>
#include <dirent.h>
#define	O_BINARY			0
#define	closesocket(a)			close(a)
#define	ERRNO				errno
#define	WINAPI
#define	LPTHREAD_START_ROUTINE		void * (*) (void *)
#endif	/* _WIN32 */

#include <sys/types.h>				/* Common */
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <stddef.h>
#include <fcntl.h>

#ifdef WITH_SSL
#include <openssl/ssl.h>
#endif /* WITH_SSL */

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

/* Macros for doubly-linked circular linked lists. */
struct llhead {
	struct llhead *prev, *next;
};
#define	LL_INIT(N)	((N)->next = (N)->prev = (N))
#define LL_HEAD(H)	struct llhead H = { &H, &H }
#define	LL_ENTRY(P,T,N)	((T *)((char *)(P) - offsetof(T, N)))
#define	LL_TAIL(H, N)							\
	do {								\
		((H)->prev)->next = (N);				\
		(N)->prev = ((H)->prev);				\
		(N)->next = (H);					\
		(H)->prev = (N);					\
	} while (0)
#define	LL_DEL(N)							\
	do {								\
		((N)->next)->prev = ((N)->prev);			\
		((N)->prev)->next = ((N)->next);			\
		LL_INIT(N);						\
	} while (0)
#define	LL_FOREACH(H,N)	for (N = (H)->next; N != (H); N = (N)->next)
#define LL_FOREACH_SAFE(H,N,T)						\
	for (N = (H)->next, T = (N)->next; N != (H);			\
			N = (T), T = (N)->next)

/* Unified socket address */
struct usa {
	unsigned			len;
	union {
		struct sockaddr		sa;
		struct sockaddr_in	sin;
	} u;
};

/* Mime type entry */
struct mimetype {
	struct mimetype	*next;
	char		*ext;			/* File extention */
	char		*mime;			/* Mime type */
	size_t		extlen;			/* Extention length */
};

/* Return values for getreqlen() function (finds HTTP request length) */
enum { REQ_INVALID = -1, REQ_INCOMPLETE = 0 };

/* Known HTTP methods */
enum { HTTP_GET, HTTP_POST, HTTP_PUT, HTTP_DELETE };

/* I/O buffer */
struct io {
	char	buf[IO_MAX];		/* Buffer */
	int	done;				/* IO finished */
	size_t	nread;				/* Bytes read */
	size_t	nwritten;			/* Bytes written */
};
#define	IO_SPACELEN(io)		(sizeof((io)->buf) - (io)->nread)
#define	IO_DATALEN(io)		((io)->nread - (io)->nwritten)

/* Connection descriptor */
struct conn {
	struct llhead	link;			/* Connections chain */
	struct usa	sa;			/* Remote socket address */
	int		sock;			/* Remote socket */
	time_t		expire;			/* Expiration time */
	time_t		ims;			/* If-Modified-Since: */
	unsigned long	clength;		/* Content-Length */
	int		reqlen;			/* Request length */
	unsigned	status;
	unsigned	http_method;		/* HTTP method */

	struct io	local;			/* Local IO buffer */
	struct io	remote;			/* Remote IO buffer */
	void (*io)(struct conn *);		/* Local IO function */
	
	char		method[METHOD_MAX + 1];	/* Used method */
	char		uri[IO_MAX + 1];	/* Requested URI */
	char		ouri[IO_MAX + 1];	/* Original unmodified URI */
	char		proto[METHOD_MAX + 1];	/* HTTP protocol  */
	
	char		*user;			/* Remote user name */
	char		*auth;			/* Authorization */
	char		*useragent;		/* User-Agent: */
	char		*path;			/* Path for get_dir */
	char		*referer;		/* Referer: */
	char		*cookie;		/* Cookie: */
	char		*ctype;			/* Content-Type: */
	char		*location;		/* Location: */
	char		*query_string;		/* GET QUERY_STRING */
	size_t		query_string_len;	/* Needed for shttpd_get_var */
	char		*pathinfo;		/* PATH_INFO */

	int		rfd;			/* Opened read file descr */
	int		wfd;			/* Opened write file descr */
	int		cgisock;		/* Socket to CGI script */
	int		dummy[2];		/* Used to fake IO */
	struct stat	st;			/* Stats of requested file */
	DIR		*dirp;			/* Opened directory */

	/* Flags */
	char		parsed;			/* Request has been parsed */
	char		done;			/* Connection to be closed */
	char		cgiparsed;		/* CGI headers parsed */

#ifdef WITH_SSL
	int		sslaccepted;		/* Flag, was accepted or not */
	SSL		*ssl;			/* SSL descriptor */
#endif /* WITH_SSL */

#ifdef _WIN32
	HANDLE		hChild;			/* Spawned CGI process */
	int		childsock;		/* Child's socket */
#endif /* _WIN32 */
};

/*
 * Globals, default values
 */
static const char	*htpass = ".htpasswd";	/* per-directory .htpasswd */
static const char	*globhtpass;		/* Global .htpasswd file */
static const char	*domain ="mydomain.com";/* Server name (auth domain) */
static const char	*ifile =
	"index.html,index.cgi";			/* Index file */
static const char	*logfile = "shttpd.log";/* Access Log file */
static const char	*errfile;		/* Error Log file */
static FILE		*logfp;			/* Log stream */
static const char	*cgipat = ".cgi";	/* CGI file pattern */
static uint16_t		lport = 80;		/* Listening port */
static char		docroot[FILENAME_MAX];	/* Working directory */
static const char	*uid;			/* Runtime User ID  */
static int		quit;			/* Exit flag */
static int		inetd;			/* Inetd mode */
static int		list_dirs = 1;		/* Do directory listing */
static int		nconns;			/* Max parallel connections */
static int		debug;			/* Show debug messages */
static int		dossl;			/* SSL mode */
static const char	*certfile ="shttpd.pem";/* SSL Certificate file */
static time_t		now;			/* Current time */
static const char	*mimefile;		/* Mime types file */
static struct mimetype	*mimetypes;		/* Known mime types */
static LL_HEAD(conns);				/* List of connections */

#ifdef WITH_SSL
static SSL_CTX		*ctx;			/* SSL context */
#endif /* WITH_SSL */

/*
 * Prototypes
 */
static void	io_inc_nwritten(struct io *io, size_t n);
static void	copypath(const char *src, char *dst, size_t dstlen);
static void	elog(int fatal, const char *fmt, ...);
static int	casecmp(register const char *s1, register const char *s2);
static int	ncasecmp(register const char *, register const char *, size_t);
static char	*mystrdup(const char *str);
static int	myopen(const char *path, int flags, int mode);
static int	mystat(const char *path, struct stat *stp);
static int	mysocketpair(int sp[2]);
static int	setdummy(struct conn *c);
static void	disconnect(struct conn *c);
static int	writeremote(struct conn *c, const char *buf, size_t len);
static int	readremote(struct conn *c, char *buf, size_t len);
static int	nonblock(int fd);
static void	newconnection(int sock, struct usa *);
static void	errtofile(const char *filename);
static int	getreqlen(const char *buf, size_t buflen);
static void	accesslog(const struct conn *c, int status);
static void	senderr(struct conn *c, int status, const char *descr,
                	const char *headers, const char *fmt, ...);
static int	montoi(const char *s, int *number);
static time_t	datetosec(const char *s);
static void	get_dir(struct conn *c);
static void	do_dir(struct conn *c);
static void	get_file(struct conn *c);
static void	do_get(struct conn *c);
static void	urldecode(char *from, char *to);
static void	killdots(char *file);
static int	getpathinfo(char *path, struct stat *stp, char **pathinfo);
static char	*fetch(const char *src, char *dst, size_t len);
static void	parseheaders(struct conn *c, char *s);
static int	useindex(char *path, size_t maxpath);
static void	decide(struct conn *c);
static void	addmimetype(const char *ext, const char *mime);
static void	freemime(void);
static void	setmimetypes(const char *file);
static void	setdefmimetypes(void);
static void	serve(struct conn *c, fd_set *rset, fd_set *wset);
int		shttpd_open_port(int port);
void		shttpd_merge_fds(fd_set *rfds, fd_set *wfds, int *maxfd);
void		shttpd_poll(int listen_socket, unsigned milliseconds);
void		shttpd_init(void);
void		shttpd_fini(void);


#ifdef EMBEDDED

#include "shttpd.h"

/* List of registered URLs */
struct userurl {
	struct llhead		link;
	const char		*url;
	shttpd_callback_t	func;
	void			*data;
};

/* List of authorization files */
struct userauth {
	struct llhead		link;
	const char		*url;
	const char		*filename;
};

static LL_HEAD(urls);		/* List head of user defined URLs */
static LL_HEAD(auths);		/* List head of user defined auth files */

static struct userurl *
isregistered(const char *url)
{
	struct llhead	*lp;
	struct userurl	*p;

	LL_FOREACH(&urls, lp) {
		p = LL_ENTRY(lp, struct userurl, link);
		if (strcmp(p->url, url) == 0)
			return (p);
	}

	return (NULL);
}

#ifdef MT
#ifndef _WIN32
#include <pthread.h>
#endif /* _WIN32 */

struct thrarg {
	struct userurl			userurl;
	struct shttpd_callback_arg	arg;
};

/*
 * This function is run in dedicated thread.
 * It executes user-defined callback function, and exits.
 * Return value is discarded.
 */
static void * WINAPI
do_thread(struct thrarg *p)
{
	int	flags;
	
	/* Put client socket to blocking mode */
#ifdef _WIN32
	flags = 0;
	(void) ioctlsocket(p->arg.connection->sock, FIONBIO, &flags);
#else
	flags = fcntl(p->arg.connection->sock, F_GETFL);
	flags &= ~O_NONBLOCK;
	(void) fcntl(p->arg.connection->sock, F_SETFL, flags);
	(void) signal(SIGPIPE, SIG_IGN);
#endif /* _WIN32 */

	/* Call user function */
	p->userurl.func(&p->arg);

	/* Free up the resources */
	disconnect(p->arg.connection);
	free(p);

	pthread_exit(0);

	return (NULL);
}
#endif /* MT */

/*
 * The URI should be handled is user-registered callback function.
 * In MT scenario, call user-defined function in dedicated thread (or process)
 * and discard the return value.
 * In nonMT scenario, call the user function. Return value should is the
 * number of bytes copied to the local IO buffer. Mark local IO as done,
 * and shttpd will take care about passing data back to client.
 */
static void
do_embedded(struct conn *conn, struct userurl *p)
{
	struct shttpd_callback_arg	arg;
#ifdef MT
	pthread_t			thr;
	struct thrarg			*param;
#endif /* MT */

	elog(0, "%s enter", __func__);

	arg.connection		= conn;
	arg.callback_data	= p->data;
	arg.buf			= conn->local.buf;
	arg.buflen		= sizeof(conn->local.buf);
	
#ifdef MT
	/* Multi-Threaded scenario. Run dedicated thread for connection. */
	if ((param = malloc(sizeof(*param))) == NULL)
		return;
	param->userurl	= *p;
	param->arg	= arg;
	
	/* FIXME check return value */
	(void) pthread_create(&thr, 0,(LPTHREAD_START_ROUTINE)do_thread, param);
	(void) pthread_detach(thr);
	LL_DEL(&conn->link);

#else
	
	/* Single-Threaded scenario. Run function and mark IO as done. */
	conn->local.nread += p->func(&arg);
	if (conn->local.nread > sizeof(conn->local.buf))
		conn->local.nread = sizeof(conn->local.buf);
	conn->local.done++;
#endif /* MT */
	
	elog(0, "%s exit", __func__);
}

#endif /* EMBEDDED */

/*
 * This function is to check the return code from snprintf() call.
 * Thanks Adam Zeldis to pointing this out in his audit report.
 */
static void
snprintf_sanitize(size_t maxbuflen, int *retval)
{
	if (*retval < -1)
		*retval = 0;
	else if (*retval > (int) maxbuflen)
		*retval = (int) maxbuflen;
}

/*
 * Increment nwritten counter. If it becomes == nread, flush both to 0
 */
static void
io_inc_nwritten(struct io *io, size_t n)
{
	assert(io->nwritten <= io->nread);
	assert(io->nread <= sizeof(io->buf));
	io->nwritten += n;
	assert(io->nwritten <= io->nread);
	if (io->nwritten == io->nread)
		io->nread = io->nwritten = 0;

}

/*
 * Change all occurences of '/' characters to OS-specific directory separator
 */
static void
copypath(const char *src, char *dst, size_t dstlen)
{
	char *end = dst + dstlen - 1;

	/* Copy the string, substituting the directory separator */
	for (; dst < end && *src != '\0'; src++, dst++)
#ifdef _WIN32
		*dst = *src == '/' ? '\\' : *src;
#else
		*dst = *src;
#endif /* _WIN32 */
	*dst = '\0';
}

#if defined(_WIN32) && !defined(__GNUC__)
static char *
getcwd(char *dst, size_t dstlen)
{
	return (GetCurrentDirectory(dstlen, dst) == 0 ? NULL : docroot);
}

/*
 * POSIX directory management (limited implementation, enough for shttpd)
 */
static int
mkdir(const char *dir, int mode)
{
	char	buf[FILENAME_MAX];

	/* Copy the string, substituting the directory separator */
	copypath(dir, buf, sizeof(buf));

	return (CreateDirectory(buf, NULL) == 0);
}

static DIR *
opendir(const char *name)
{
	DIR		*dir = NULL;
	size_t		base_length;
	const char	*all;

	if (name && name[0]) {
		base_length = strlen(name);
		all = strchr("/\\", name[base_length - 1]) ? "*" : "/*";

		if ((dir = malloc(sizeof *dir)) != NULL &&
		    (dir->name = malloc(base_length + strlen(all) + 1)) != 0) {
			(void) strcat(strcpy(dir->name, name), all);

			if ((dir->handle = (long) _findfirst(dir->name,
			    &dir->info)) != -1) {
				dir->result.d_name = 0;
			} else {
				free(dir->name);
				free(dir);
				dir = 0;
			}
		} else {
			free(dir);
			dir = NULL;
			errno = ENOMEM;
		}
	} else {
		errno = EINVAL;
	}

	return (dir);
}

static int
closedir(DIR *dir)
{
	int result = -1;

	if (dir) {
		if(dir->handle != -1)
			result = _findclose(dir->handle);

		free(dir->name);
		free(dir);
	}

	if (result == -1) 
		errno = EBADF;

	return (result);
}

static struct dirent *
readdir(DIR *dir)
{
	struct dirent *result = 0;

	if (dir && dir->handle != -1) {
		if(!dir->result.d_name ||
		    _findnext(dir->handle, &dir->info) != -1) {
			result = &dir->result;
			result->d_name = dir->info.name;
		}
	} else {
		errno = EBADF;
	}

	return (result);
}
#endif /* _WIN32 */

/*
 * Log function
 */
static void
elog(int fatal, const char *fmt, ...)
{
	va_list	ap;

	if (debug == 0 && fatal == 0)
		return;

	(void) fprintf(stderr, "%lu ", (unsigned long) now);

	va_start(ap, fmt);
	(void) vfprintf(stderr, fmt, ap);
	va_end(ap);

	(void) fputc('\n', stderr);
	(void) fflush(stderr);

	if (fatal)
		exit(EXIT_FAILURE);
}

/*
 * Case-insensitive string comparison, a-la strcmp()
 */
static int
casecmp(register const char *s1, register const char *s2)
{
	for (; *s1 != '\0' && *s2 != '\0'; s1++, s2++)
		if (tolower(*s1) != tolower(*s2))
			break;
	
	return (*s1 - *s2);
}

/* Case insensitive memory comparison, strncasecmp() */
static int
ncasecmp(register const char *s1, register const char *s2, size_t len)
{
	register const char	*e = s1 + len - 1;
	int			ret;
	
	for (; s1 < e && *s1 != '\0' && *s2 != '\0' &&
	    tolower(*s1) == tolower(*s2); s1++, s2++) ;
	ret = tolower(*s1) - tolower(*s2);

	return (ret);
}

/* strdup() is not standard, define it here */
static char *
mystrdup(const char *str)
{
	int	len;
	char	*p;

	len = strlen(str);
	if ((p = malloc(len + 1)) != NULL)
		(void) strcpy(p, str);

	return (p);
}

/*
 * Wrapper around open(), that takes care about directory separators
 */
static int
myopen(const char *path, int flags, int mode)
{
#ifdef _WIN32
	char	buf[FILENAME_MAX];
#endif /* _WIN32 */
	int	fd;
	
	elog(0, "%s enter [%s] %d %d", __func__, path, flags, mode);
	
#ifdef _WIN32
	copypath(path, buf, sizeof(buf));
	fd = open(buf, flags, mode);
#else
	fd = open(path, flags, mode);
	(void) fcntl(fd, F_SETFD, FD_CLOEXEC);
#endif /* _WIN32 */
	
	elog(0, "%s exit %d err %d", __func__, fd, ERRNO);
	
	return (fd);
}

/*
 * The wrapper around stat(), that takes care about directory separators
 */
static int
mystat(const char *path, struct stat *stp)
{
#ifdef _WIN32
	char	buf[FILENAME_MAX];
	int	len;
#endif /* _WIN32  */
	int	ret;
	
	elog(0, "%s enter [%s]", __func__, path);

#ifdef _WIN32
	copypath(path, buf, sizeof(buf));
	len = strlen(buf);
	if (buf[len - 1] == '\\')
		buf[len - 1] = '\0';
	ret = stat(buf, stp);
#else
	ret = stat(path, stp);
#endif /* _WIN32 */

	elog(0, "%s exit %d, err %d", __func__, ret, ERRNO);
	
	return (ret);
}

/*
 * Create a pair of connected sockets, that can be used a duplex data channel.
 * Return 0 on success, -1 on error
 */
static int
mysocketpair(int sp[2])
{
	struct sockaddr_in	sa;
	int			sock, ret = -1;
	unsigned		len = sizeof(sa);

	(void) memset(&sa, 0, sizeof(sa));
	sa.sin_family 		= AF_INET;
	sa.sin_port		= htons(0);
	sa.sin_addr.s_addr	= htonl(INADDR_LOOPBACK);

	elog(0, "%s enter", __func__);
	if ((sock = socket(AF_INET, SOCK_STREAM, 6)) == -1) {
		elog(0, "mysocketpair: socket(): %d", ERRNO);
	} else if (bind(sock, (struct sockaddr *) &sa, len) != 0) {
		elog(0, "mysocketpair: bind(): %d", ERRNO);
		(void) closesocket(sock);
	} else if (listen(sock, 1) != 0) {
		elog(0, "mysocketpair: listen(): %d", ERRNO);
		(void) closesocket(sock);
	} else if (getsockname(sock, (struct sockaddr *) &sa, &len) != 0) {
		elog(0, "mysocketpair: getsockname(): %d", ERRNO);
		(void) closesocket(sock);
	} else if ((sp[0] = socket(AF_INET, SOCK_STREAM, 6)) == -1) {
		elog(0, "mysocketpair: socket(): %d", ERRNO);
		(void) closesocket(sock);
	} else if (connect(sp[0], (struct sockaddr *) &sa, len) != 0) {
		elog(0, "mysocketpair: connect(): %d", ERRNO);
		(void) closesocket(sock);
		(void) closesocket(sp[0]);
	} else if ((sp[1] = accept(sock,(struct sockaddr *) &sa, &len)) == -1) {
		elog(0, "mysocketpair: accept(): %d", ERRNO);
		(void) closesocket(sock);
		(void) closesocket(sp[0]);
	} else {
		/* Success */
		ret = 0;
		(void) closesocket(sock);
	}
	elog(0, "%s exit", __func__);

	return (ret);
}

/*
 * shttpd does select() for multiplex IO. But some channels do not
 * have opened file descriptors, like opened directory or user-defined
 * function. On win32 we cannot even pass regular file descriptor to
 * the select(). Therefore, we do the following:
 *   o in such a connection, create a socket pair.
 *   o write one byte of data into pair[0]
 *   o pass pair[0] to the select(), so this connection will be always
 *     marked 'active'.
 * Return values: 0 on success, -1 on error
 */
static int
setdummy(struct conn *c)
{
	int	ret = -1;

	if (mysocketpair(c->dummy) == 0) {
		(void) send(c->dummy[1], "", 1, 0);
		ret = 0;
	}

	return (ret);
}

/*
 * Disconnect from remote side, free resources
 */
static void
disconnect(struct conn *c)
{
	elog(0, "%s enter [%p]", __func__, c);
	LL_DEL(&c->link);

	/* If parseheaders() allocated any data, free it */
	free(c->useragent);
	free(c->user);
	free(c->cookie);
	free(c->ctype);
	free(c->referer);
	free(c->location);
	free(c->auth);
	free(c->path);

	/* Free resources */
	if (c->dummy[0] != -1)	(void) closesocket(c->dummy[0]);
	if (c->dummy[1] != -1)	(void) closesocket(c->dummy[1]);
	if (c->rfd != -1)	(void) close(c->rfd);
	if (c->wfd != -1)	(void) close(c->wfd);
	if (c->cgisock != -1)	(void) closesocket(c->cgisock);
	if (c->dirp)		(void) closedir(c->dirp);

#ifdef WITH_SSL
	if (c->ssl)
		SSL_free(c->ssl);
#endif /* WITH_SSL */
	
	(void) closesocket(c->sock);
	free(c);

	/* In inetd mode, exit if request is finished. */
	if (inetd)
		quit++;

	elog(0, "%s exit", __func__);
}

#ifdef WITH_SSL
/*
 * Perform SSL handshake
 */
static void
handshake(struct conn *c)
{
	int	n;

	elog(0, "%s enter", __func__);
	if ((n = SSL_accept(c->ssl)) == 0) {
		n = SSL_get_error(c->ssl, n);
		if (n != SSL_ERROR_WANT_READ && n != SSL_ERROR_WANT_WRITE)
			disconnect(c);
		elog(0, "handshake: SSL_accept error %d", n);
	} else {
		elog(0, "handshake: SSL accepted");
		c->sslaccepted = 1;
	}
	elog(0, "%s exit", __func__);
}
#endif /* WITH_SSL */

/*
 * Send data to a remote end. Return bytes sent.
 */
static int
writeremote(struct conn *c, const char *buf, size_t len)
{
	int	n;

	elog(0, "%s enter %p %u", __func__, c, len);
#ifdef WITH_SSL
	if (c->ssl)
		n = SSL_write(c->ssl, buf, len);
	else
#endif /* WITH_SSL */
		n = send(c->sock, buf, len, 0);
	if (n  == 0 || (n < 0 && ERRNO != EWOULDBLOCK))
		c->remote.done = 1;
	elog(0, "%s exit %p: %d written, err %d", __func__, c, n, ERRNO);
		
	return (n);
}

/*
 * Read data from the remote end. Return bytes read.
 */
static int
readremote(struct conn *c, char *buf, size_t len)
{
	int	n = -1;
	elog(0, "%s enter", __func__);
#ifdef WITH_SSL
	if (c->ssl && c->sslaccepted == 0)
		handshake(c);
	else if (c->ssl)
		n = SSL_read(c->ssl, buf, len);
	else
#endif /* WITH_SSL*/
		n = recv(c->sock, buf, len, 0);
	
	if (n == 0 || (n < 0 && ERRNO != EWOULDBLOCK))
		c->remote.done = 1;
	elog(0, "readremote(%p): %d bytes of %u read [%d]", c, n, len, ERRNO);
	elog(0, "%s exit", __func__);

	return (n);
}

/*
 * Put given file descriptor in blocking (block == 1)
 * or non-blocking (block == 0) mode.
 * Return 0 if success, or -1
 */
static int
nonblock(int fd)
{
	int	ret = -1;
#ifdef	_WIN32
	unsigned long	on = 1;

	return (ioctlsocket(fd, FIONBIO, &on));
#else
	int	flags;

	elog(0, "%s enter", __func__);
	if ((flags = fcntl(fd, F_GETFL, 0)) == -1)
		elog(0, "nonblock: fcntl(F_GETFL): %d", ERRNO);
	else if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) != 0)
		elog(0, "nonblock: fcntl(F_SETFL): %d", ERRNO);
	else
		ret = 0;	/* Success */
	elog(0, "%s exit", __func__);

	return (ret);
#endif /* _WIN32 */
}

/*
 * Setup listening socket on given port, return socket
 */
int
shttpd_open_port(int port)
{
	int		sock, on = 1;
	struct usa	sa;

	sa.len				= sizeof(sa.u.sin);
	sa.u.sin.sin_family		= AF_INET;
	sa.u.sin.sin_port		= htons((uint16_t) port);
	sa.u.sin.sin_addr.s_addr	= htonl(INADDR_ANY);

	elog(0, "%s enter", __func__);
	if ((sock = socket(PF_INET, SOCK_STREAM, 6)) == -1)
		elog(1, "shttpd_open_port: socket: %s", strerror(ERRNO));
	else if (nonblock(sock) != 0)
		elog(1, "shttpd_open_port: nonblock");
	else if (setsockopt(sock, SOL_SOCKET,
	    SO_REUSEADDR,(char *) &on, sizeof(on)) != 0)
		elog(1, "shttpd_open_port: setsockopt");
	else if (bind(sock, &sa.u.sa, sa.len) < 0)
		elog(1, "shttpd_open_port: bind(%d): %s", port, strerror(ERRNO));
	else if (listen(sock, 128) != 0)
		elog(1, "shttpd_open_port: listen: %s", strerror(ERRNO));
#ifndef _WIN32
	(void) fcntl(sock, F_SETFD, FD_CLOEXEC);
#endif /* !_WIN32 */
	elog(0, "%s exit", __func__);

	return (sock);
}

/*
 * newconnection new connection
 */
static void
newconnection(int sock, struct usa *usa) {
	struct conn	*c;
	struct usa	sa;
#ifdef WITH_SSL
	SSL		*ssl = NULL;
#endif /* WITH_SSL */
			
	sa.len = sizeof(sa.u.sin);

	if (nonblock(sock) != 0) {
		elog(0, "newconnection: nonblock: %s", strerror(ERRNO));
		(void) close(sock);
#ifdef WITH_SSL
	} else if (dossl && (ssl = SSL_new(ctx)) == NULL) {
		elog(0, "newconnection: SSL_new: %s", strerror(ERRNO));
		(void) close(sock);
	} else if (dossl && SSL_set_fd(ssl, sock) == 0) {
		elog(0, "newconnection: SSL_set_fd: %s", strerror(ERRNO));
		(void) close(sock);
		SSL_free(ssl);
#endif /* WITH_SSL */
	} else if ((c = calloc(1, sizeof(*c))) == NULL) {
		(void) close(sock);
		elog(0, "newconnection: calloc: %s", strerror(ERRNO));
	} else {
		LL_TAIL(&conns, &c->link);
		if (usa) {
			c->sa = *usa;
		} else {
			(void) getpeername(sock, &sa.u.sa, &sa.len);
			c->sa = sa;
		}
		c->sock = sock;
		
		c->cgisock = c->rfd = c->wfd = c->dummy[0] = c->dummy[1] = -1;
#ifndef _WIN32
		(void) fcntl(sock, F_SETFD, FD_CLOEXEC);
#endif /* _WIN32 */

#ifdef WITH_SSL
		c->ssl = ssl;
		if (ssl)
			handshake(c);
#endif /* WITH_SSL */
		
		c->expire = now + EXPIRE_TIME;
		elog(0, "%p %d connected.", c, sock);
	}
}

/*
 * Redirect error log to a file
 */
static void
errtofile(const char *filename)
{
	int fd;
	
	elog(0, "%s enter", __func__);
	if ((fd = myopen(filename, O_WRONLY | O_APPEND | O_CREAT, 0666)) == -1)
		elog(1, "open(%s): %s", errfile, strerror(ERRNO));

	(void) dup2(fd, 2);
	(void) close(fd);
	elog(0, "%s exit", __func__);
}

/*
 * Check whether full request is buffered
 * Return request length, or 0
 */
static int
getreqlen(const char *buf, size_t buflen)
{
	const char	*s, *e;
	int		len = REQ_INCOMPLETE;

	elog(0, "%s enter", __func__);
	for (s = buf, e = s + buflen - 1; len == REQ_INCOMPLETE && s < e; s++)
		if (!isprint(*(unsigned char *) s) && *s != '\r' && *s != '\n')
			len = REQ_INVALID;
		else if (s[0] == '\n' && s[1] == '\n')
			len = s - buf + 2;
		else if (s[0] == '\n' && &s[1] < e &&
		    s[1] == '\r' && s[2] == '\n')
			len = s - buf + 3;
	elog(0, "%s exit [%d]", __func__, len);

	return (len);
}

/*
 * Write an HTTP access log into a file `logfile'
 */
static void
accesslog(const struct conn *c, int status)
{
	char	date[64];

	if (logfp == NULL)
		return;
	
	elog(0, "%s enter", __func__);
	
	(void) strftime(date, sizeof(date),"%d/%b/%Y %H:%M:%S",localtime(&now));
	(void) fprintf(logfp, "%s - %s [%s] \"%s %s %s\" %d\n",
	    inet_ntoa(c->sa.u.sin.sin_addr), c->user ? c->user : "-",
	    date, c->method, c->uri, c->proto, status);
	(void) fflush(logfp);
	
	elog(0, "%s exit", __func__);
}

/*
 * Send an error back to a client.
 */
static void
senderr(struct conn *c, int status, const char *descr,
		const char *headers, const char *fmt, ...)
{
	va_list	ap;
	int	n;

	elog(0, "%s enter", __func__);
	
	n = snprintf(c->local.buf, sizeof(c->local.buf),
	    "HTTP/1.1 %d %s\r\n%s%s\r\n%d ",
	    status, descr, headers, headers[0] == '\0' ? "" : "\r\n", status);
	snprintf_sanitize(sizeof(c->local.buf), &n);
	va_start(ap, fmt);
	n += vsnprintf(c->local.buf + n, sizeof(c->local.buf) - n, fmt, ap);
	snprintf_sanitize(sizeof(c->local.buf), &n);
	va_end(ap);
	c->local.nread = n;
	accesslog(c, status);
	c->local.done++;
	
	elog(0, "%s exit", __func__);
}

/*
 * Convert month to the month number. Return 1 if OK, 0 if error
 */
static int
montoi(const char *s, int *number)
{
	int	num = -1;
	
	elog(0, "%s enter", __func__);
	
	if (!casecmp(s, "Jan") || !casecmp(s, "January"))
		num = 0;
	else if (!casecmp(s, "Feb") || !casecmp(s, "February"))
		num = 1;
	else if (!casecmp(s, "Mar") || !casecmp(s, "March"))
		num = 2;
	else if (!casecmp(s, "Apr") || !casecmp(s, "April"))
		num = 3;
	else if (!casecmp(s, "May"))
		num = 4;
	else if (!casecmp(s, "Jun") || !casecmp(s, "June"))
		num = 5;
	else if (!casecmp(s, "Jul") || !casecmp(s, "July"))
		num = 6;
	else if (!casecmp(s, "Aug") || !casecmp(s, "August"))
		num = 7;
	else if (!casecmp(s, "Sep") || !casecmp(s, "September"))
		num = 8;
	else if (!casecmp(s, "Oct") || !casecmp(s, "October"))
		num = 9;
	else if (!casecmp(s, "Nov") || !casecmp(s, "November"))
		num = 10;
	else if (!casecmp(s, "Dec") || !casecmp(s, "December"))
		num = 11;

	*number = num;
	
	elog(0, "%s exit", __func__);
	
	return (num != -1);
}

/*
 * Parse date-time string, and return the corresponding time_t value
 */
static time_t
datetosec(const char *s)
{
	struct tm	tm;
	char		mon[32];
	int		sec, min, hour, mday, month, year;

	elog(0, "%s enter", __func__);
	(void) memset(&tm, 0, sizeof(tm));

	/* DD/month/YY HH:MM:SS */
	if (((sscanf(s, "%d/%32[a-zA-Z]/%d %d:%d:%d",
	    &mday, mon, &year, &hour, &min, &sec) == 6) ||
	    (sscanf(s, "%d %32[a-zA-Z] %d %d:%d:%d",
	    &mday, mon, &year, &hour, &min, &sec) == 6) ||
	    (sscanf(s, "%d-%32[a-zA-Z]-%d %d:%d:%d",
	    &mday, mon, &year, &hour, &min, &sec) == 6))
	    && montoi(mon, &month)) {
		tm.tm_mday	= mday;
		tm.tm_mon	= month;
		tm.tm_year	= year;
		tm.tm_hour	= hour;
		tm.tm_min	= min;
		tm.tm_sec	= sec;
	}
	
	if (tm.tm_year > 1900)
		tm.tm_year -= 1900;
	else if (tm.tm_year < 70)
		tm.tm_year += 100;

	elog(0, "datetosec: \"%s\" == %u", s, mktime(&tm));
	elog(0, "%s exit", __func__);

	return (mktime(&tm));
}

#ifdef WITH_PUT_AND_DELETE
/*
 * For a given PUT path, create all intermediate subdirectories
 * for given path. Return 0 if the path itself is a directory,
 * or -1 on error, 1 if OK.
 */
static int
put_dir(const char *path)
{
	char		buf[FILENAME_MAX];
	const char	*s, *p;
	struct stat	st;
	size_t		len;

	elog(0, "%s enter", __func__);
	for (s = p = path + 2; (p = strchr(s, '/')) != NULL; s = ++p) {
		len = p - path;
		assert(len < sizeof(buf));
		(void) memcpy(buf, path, len);
		buf[len] = '\0';

		/* Try to create intermediate directory */
		if (mystat(buf, &st) == -1 && mkdir(buf, 0755) != 0)
			return (-1);

		/* Is path itself a directory ? */
		if (p[1] == '\0')
			return (0);
	}
	elog(0, "%s exit", __func__);

	return (1);
}

/*
 * PUT request
 */
static void
put_file(struct conn *c)
{
	int	n, len;

	elog(0, "%s enter", __func__);
	assert(c->wfd != -1);
	if ((len = IO_DATALEN(&c->remote)) <= 0)
		return;

	n = write(c->wfd, c->remote.buf + c->remote.nwritten, len);
	elog(0, "put_file(%p): %d bytes", c, n);
	
	if (n > 0) {
		io_inc_nwritten(&c->remote, n);
	} else {
		c->local.done++;
		c->io = NULL;
	}
	elog(0, "%s exit", __func__);
}
#endif /* WITH_PUT_AND_DELETE */

/*
 * GET the directory
 */
static void
get_dir(struct conn *c)
{
	struct dirent	*dp = NULL;
	char		file[FILENAME_MAX], line[FILENAME_MAX + 512],
				size[64], mod[64];
	const char	*slash;
	struct stat	st;
	int		n, left;

	elog(0, "%s enter", __func__);
	assert(c->dirp != NULL);
	assert(c->uri[0] != '\0');

	left = IO_SPACELEN(&c->local);
	slash = c->uri[strlen(c->uri) - 1] == '/' ? "" : "/";

	do {
		if (left < (int) sizeof(line))
			break;

		if ((dp = readdir(c->dirp)) == NULL) {
			/* Finished reading directory */
			c->local.done++;
			c->io = NULL;
			break;
		}
		
		/* Do not show current dir and passwords file */
		if (!strcmp(dp->d_name, ".") || !strcmp(dp->d_name, htpass))
			continue;
		
		(void) snprintf(file, sizeof(file),
		    "%s%s%s",c->path, slash, dp->d_name);
		(void) mystat(file, &st);
		if (st.st_mode & S_IFDIR)
			snprintf(size,sizeof(size),"        %s","&lt;DIR&gt;");
		else
			(void) snprintf(size, sizeof(size),"%10.2f kB",
			    (float) st.st_size / 1024);
		(void) strftime(mod, sizeof(mod),
		    "%d-%b-%Y %H:%M", localtime(&st.st_mtime));

		n = snprintf(line, sizeof(line),
		    "<tr><td><a href=\"%s%s%s\">%s%s</a></td>"
		    "<td>&nbsp;&nbsp;%s</td>"
		    "<td><pre>%s</pre></td></tr>\n",
		    c->uri, slash, dp->d_name, dp->d_name,
		    st.st_mode & S_IFDIR ? "/" : "", mod, size);
		snprintf_sanitize(sizeof(line), &n);
		(void) memcpy(c->local.buf + c->local.nread, line, n);
		c->local.nread += n;
		left -= n;
	} while (dp != NULL);
	elog(0, "%s exit", __func__);
}

/*
 * Schedule GET for the directory
 */
static void
do_dir(struct conn *c)
{
	elog(0, "do_dir");

	elog(0, "%s enter [%s]", __func__, c->path);
	if ((c->dirp = opendir(c->path)) == NULL) {
		accesslog(c, 500);
		senderr(c, 500, "Error","", "Cannot open dir");
	} else if (setdummy(c) != 0) {
		accesslog(c, 500);
		senderr(c, 500, "Error","", "Cannot open dir (dummy)");
	} else {
		c->local.nread = snprintf(c->local.buf, sizeof(c->local.buf),
		    "HTTP/1.1 200 OK\r\n"
		    "Content-Type: text/html\r\n"
		    "\r\n"
		    "<html><head><title>Index of %s</title></head>"
		    "<body><h1>Index of %s</h1><pre><table>"
		    "<tr><th align=\"left\">Name</th>"
		    "<th align=\"left\">&nbsp;&nbsp;Modified</th>"
		    "<th align=\"right\">Size</th></tr>",
		    c->uri, c->uri);
		snprintf_sanitize(sizeof(c->local.buf), (int *)&c->local.nread);
		c->io = get_dir;
		c->remote.nread = 0;
		accesslog(c, 200);
	}
	elog(0, "%s exit", __func__);
}

/*
 * GET regular file
 */
static void
get_file(struct conn *c)
{
	int	n;

	elog(0, "%s enter %p", __func__, c);
	assert(c->rfd != -1);
	n = read(c->rfd, c->local.buf + c->local.nread, IO_SPACELEN(&c->local));
	
	if (n > 0) {
		c->local.nread += n;
	} else {
		c->local.done++;
		c->io = NULL;
	}
	elog(0, "%s exit, %p read %d", __func__, c, n);
}

/*
 * Schedule GET for regular file
 */
static void
do_get(struct conn *c)
{
	char		date[64], lm[64], etag[64];
	const char	*mime = "text/html";		/* Default mime type */
	const char	*s = c->uri + strlen(c->uri);
	struct mimetype	*p;

	elog(0, "%s enter [%s]", __func__, c->uri);

#ifdef _WIN32
	if (setdummy(c) != 0) {
		senderr(c, 500, "Error", "", "Error opening dummy file");
		return;
	}
#endif /* _WIN32 */
	
	/* Figure out the mime type */
	for (p = mimetypes; p != NULL; p = p->next)
		if (strlen(c->uri) > p->extlen &&
		    *(s - p->extlen - 1) == '.' &&
		    !ncasecmp(p->ext, s - p->extlen, p->extlen)) {
			mime = p->mime;
			break;
		}

	/* Prepare Etag, Date, Last-Modified headers */
	strftime(date, sizeof(date), "%d/%b/%Y %H:%M:%S", localtime(&now));
	(void) strftime(lm, sizeof(lm), "%d/%b/%Y %H:%M:%S",
	    localtime(&c->st.st_mtime));
	(void) snprintf(etag, sizeof(etag), "%lx.%lx",
	    (unsigned long) c->st.st_mtime, (unsigned long) c->st.st_size);

	/* Local read buffer should be empty */
	c->local.nread = snprintf(c->local.buf, sizeof(c->local.buf),
	    "HTTP/1.1 200 OK\r\n"
	    "Connection: close\r\n"
	    "Date: %s\r\n"
	    "Last-Modified: %s\r\n"
	    "Content-Type: %s\r\n"
	    "Content-Length: %lu\r\n"
	    "\r\n", date, lm, mime, (unsigned long) c->st.st_size);
	snprintf_sanitize(sizeof(c->local.buf), (int *) &c->local.nread);
	c->io = get_file;
	c->remote.nread = 0;
	elog(0, "%s exit", __func__);
}

/*
 * Decode urlencoded string. from and to may point to the same location
 */
static void
urldecode(char *from, char *to)
{
	int	i, a, b;
#define	HEXTOI(x)  (isdigit(x) ? x - '0' : x - 'W')

	elog(0, "%s enter [%s]", __func__, from);
	for (i = 0; *from != '\0'; i++, from++) {
		if (from[0] == '%' &&
		    isxdigit((unsigned char) from[1]) &&
		    isxdigit((unsigned char) from[2])) {
			a = tolower(from[1]);
			b = tolower(from[2]);
			to[i] = (HEXTOI(a) << 4) | HEXTOI(b);
			from += 2;
		} else {
			to[i] = *from;
		}
	}
	to[i] = '\0';
	elog(0, "%s exit [%s]", __func__, to);
}

/*
 * Protect from ..
 */
static void
killdots(char *file)
{
	char	good[FILENAME_MAX], *s;
	size_t	i;

	elog(0, "%s enter %s", __func__, file);

#define	BADCHAR(x)	((x) == '.')

	/* Get rid of dots */
	for (s = file, i = 0; *s != '\0'; s++) {
		good[i++] = *s;
		while (BADCHAR(s[0]) && BADCHAR(s[1]))
			s++;
	}
	assert(i < sizeof(good));
	good[i] = '\0';
	(void) strcpy(file, good);
	
	elog(0, "%s exit %s", __func__, file);
}

/*
 * stat() requested file, return 0 if OK, -1 if error.
 * If the file is given arguments using PATH_INFO mechanism,
 * initialize pathinfo pointer.
 */
static int
getpathinfo(char *path, struct stat *stp, char **pathinfo)
{
	char	*p;
	int	ret = -1;

	elog(0, "%s enter %s", __func__, path);
	*pathinfo = NULL;

	if (mystat(path, stp) == 0) {
		ret = 0;
	} else {
		/* Strip directory parts of the path one by one */
		for (p = path + strlen(path) - 1; p > path + 2; p--)
			if (*p == '/') {
				*p = '\0';
				if (mystat(path, stp) == 0) {
					*pathinfo = p + 1;
					ret = 0;
					break;
				} else {
					*p = '/';
				}
			}
	}
	elog(0, "%s exit %d", __func__, ret);

	return (ret);
}

/*
 * Fetch header value into specified destination
 */
static char *
fetch(const char *src, char *dst, size_t len)
{
	const char	*p;

	elog(0, "%s enter", __func__);
	*dst = '\0';
	if ((p = strchr(src, '\r')) || (p = strchr(src, '\n'))) {
		len--;	/* For terminating '\0' */
		if (p - src < (int) len)
			len = p - src;
		(void) memcpy(dst, src, len);
		dst[len] = '\0';
	}
	elog(0, "%s exit", __func__);

	return (dst);
}

/*
 * Parse HTTP headers, filling values in struct conn
 */
static void
parseheaders(struct conn *c, char *s)
{
	char	val[IO_MAX];

	elog(0, "%s enter", __func__);
	while (s != NULL && s[0] != '\n' && s[0] != '\r') {
		if (ncasecmp(s, "User-Agent: ", 12) == 0)
			c->useragent = mystrdup(fetch(s + 12, val,sizeof(val)));
		else if (ncasecmp(s, "Content-Type: ", 14) == 0)
			c->ctype = mystrdup(fetch(s + 14, val, sizeof(val)));
		else if (ncasecmp(s, "Content-Length: ", 16) == 0)
			c->clength = strtoul(fetch(s + 15,
			    val, sizeof(val)), NULL, 10);
		else if (ncasecmp(s, "If-Modified-Since: ", 19) == 0)
			c->ims = datetosec(fetch(s + 19, val, sizeof(val)));
		else if (ncasecmp(s, "Authorization: ", 15) == 0)
			c->auth = mystrdup(fetch(s + 15, val, sizeof(val)));
		else if (ncasecmp(s, "Referer: ", 9) == 0)
			c->referer = mystrdup(fetch(s + 9, val, sizeof(val)));
		else if (ncasecmp(s, "Cookie: ", 8) == 0)
			c->cookie = mystrdup(fetch(s + 8, val, sizeof(val)));
		else if (ncasecmp(s, "Location: ", 10) == 0)
			c->location = mystrdup(fetch(s + 10, val, sizeof(val)));
		else if (ncasecmp(s, "Status: ", 8) == 0)
			c->status = strtoul(fetch(s+8, val, sizeof(val)),0,10);
	
		if ((s = strchr(s, '\n')) != NULL)
			s++;
	}
	elog(0, "%s exit", __func__);
}

#ifndef NO_AUTH
/*
 * MD5 crypto algorithm.
 */
typedef struct _MD5Context {
	uint32_t	buf[4];
	uint32_t	bits[2];
	unsigned char	in[64];
} _MD5_CTX;

#if __BYTE_ORDER == 1234
#define byteReverse(buf, len)	/* Nothing */
#else
/*
 * Note: this code is harmless on little-endian machines.
 */
static void byteReverse(unsigned char *buf, unsigned longs)
{
	uint32_t t;
	do {
		t = (uint32_t) ((unsigned) buf[3] << 8 | buf[2]) << 16 |
			((unsigned) buf[1] << 8 | buf[0]);
		*(uint32_t *) buf = t;
		buf += 4;
	} while (--longs);
}
#endif /* __BYTE_ORDER */


/* The four core functions - F1 is optimized somewhat */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (x ^ y ^ z)
#define F4(x, y, z) (y ^ (x | ~z))

/* This is the central step in the MD5 algorithm. */
#define MD5STEP(f, w, x, y, z, data, s) \
( w += f(x, y, z) + data,  w = w<<s | w>>(32-s),  w += x )

/*
 * Start MD5 accumulation.  Set bit count to 0 and buffer to mysterious
 * initialization constants.
 */
static void _MD5Init(_MD5_CTX *ctx)
{
	ctx->buf[0] = 0x67452301;
	ctx->buf[1] = 0xefcdab89;
	ctx->buf[2] = 0x98badcfe;
	ctx->buf[3] = 0x10325476;

	ctx->bits[0] = 0;
	ctx->bits[1] = 0;
}


/*
 * The core of the MD5 algorithm, this alters an existing MD5 hash to
 * reflect the addition of 16 longwords of new data.  MD5Update blocks
 * the data and converts bytes into longwords for this routine.
 */
static void _MD5Transform(uint32_t buf[4], uint32_t const in[16])
{
	register uint32_t a, b, c, d;

	a = buf[0];
	b = buf[1];
	c = buf[2];
	d = buf[3];

	MD5STEP(F1, a, b, c, d, in[0] + 0xd76aa478, 7);
	MD5STEP(F1, d, a, b, c, in[1] + 0xe8c7b756, 12);
	MD5STEP(F1, c, d, a, b, in[2] + 0x242070db, 17);
	MD5STEP(F1, b, c, d, a, in[3] + 0xc1bdceee, 22);
	MD5STEP(F1, a, b, c, d, in[4] + 0xf57c0faf, 7);
	MD5STEP(F1, d, a, b, c, in[5] + 0x4787c62a, 12);
	MD5STEP(F1, c, d, a, b, in[6] + 0xa8304613, 17);
	MD5STEP(F1, b, c, d, a, in[7] + 0xfd469501, 22);
	MD5STEP(F1, a, b, c, d, in[8] + 0x698098d8, 7);
	MD5STEP(F1, d, a, b, c, in[9] + 0x8b44f7af, 12);
	MD5STEP(F1, c, d, a, b, in[10] + 0xffff5bb1, 17);
	MD5STEP(F1, b, c, d, a, in[11] + 0x895cd7be, 22);
	MD5STEP(F1, a, b, c, d, in[12] + 0x6b901122, 7);
	MD5STEP(F1, d, a, b, c, in[13] + 0xfd987193, 12);
	MD5STEP(F1, c, d, a, b, in[14] + 0xa679438e, 17);
	MD5STEP(F1, b, c, d, a, in[15] + 0x49b40821, 22);

	MD5STEP(F2, a, b, c, d, in[1] + 0xf61e2562, 5);
	MD5STEP(F2, d, a, b, c, in[6] + 0xc040b340, 9);
	MD5STEP(F2, c, d, a, b, in[11] + 0x265e5a51, 14);
	MD5STEP(F2, b, c, d, a, in[0] + 0xe9b6c7aa, 20);
	MD5STEP(F2, a, b, c, d, in[5] + 0xd62f105d, 5);
	MD5STEP(F2, d, a, b, c, in[10] + 0x02441453, 9);
	MD5STEP(F2, c, d, a, b, in[15] + 0xd8a1e681, 14);
	MD5STEP(F2, b, c, d, a, in[4] + 0xe7d3fbc8, 20);
	MD5STEP(F2, a, b, c, d, in[9] + 0x21e1cde6, 5);
	MD5STEP(F2, d, a, b, c, in[14] + 0xc33707d6, 9);
	MD5STEP(F2, c, d, a, b, in[3] + 0xf4d50d87, 14);
	MD5STEP(F2, b, c, d, a, in[8] + 0x455a14ed, 20);
	MD5STEP(F2, a, b, c, d, in[13] + 0xa9e3e905, 5);
	MD5STEP(F2, d, a, b, c, in[2] + 0xfcefa3f8, 9);
	MD5STEP(F2, c, d, a, b, in[7] + 0x676f02d9, 14);
	MD5STEP(F2, b, c, d, a, in[12] + 0x8d2a4c8a, 20);

	MD5STEP(F3, a, b, c, d, in[5] + 0xfffa3942, 4);
	MD5STEP(F3, d, a, b, c, in[8] + 0x8771f681, 11);
	MD5STEP(F3, c, d, a, b, in[11] + 0x6d9d6122, 16);
	MD5STEP(F3, b, c, d, a, in[14] + 0xfde5380c, 23);
	MD5STEP(F3, a, b, c, d, in[1] + 0xa4beea44, 4);
	MD5STEP(F3, d, a, b, c, in[4] + 0x4bdecfa9, 11);
	MD5STEP(F3, c, d, a, b, in[7] + 0xf6bb4b60, 16);
	MD5STEP(F3, b, c, d, a, in[10] + 0xbebfbc70, 23);
	MD5STEP(F3, a, b, c, d, in[13] + 0x289b7ec6, 4);
	MD5STEP(F3, d, a, b, c, in[0] + 0xeaa127fa, 11);
	MD5STEP(F3, c, d, a, b, in[3] + 0xd4ef3085, 16);
	MD5STEP(F3, b, c, d, a, in[6] + 0x04881d05, 23);
	MD5STEP(F3, a, b, c, d, in[9] + 0xd9d4d039, 4);
	MD5STEP(F3, d, a, b, c, in[12] + 0xe6db99e5, 11);
	MD5STEP(F3, c, d, a, b, in[15] + 0x1fa27cf8, 16);
	MD5STEP(F3, b, c, d, a, in[2] + 0xc4ac5665, 23);

	MD5STEP(F4, a, b, c, d, in[0] + 0xf4292244, 6);
	MD5STEP(F4, d, a, b, c, in[7] + 0x432aff97, 10);
	MD5STEP(F4, c, d, a, b, in[14] + 0xab9423a7, 15);
	MD5STEP(F4, b, c, d, a, in[5] + 0xfc93a039, 21);
	MD5STEP(F4, a, b, c, d, in[12] + 0x655b59c3, 6);
	MD5STEP(F4, d, a, b, c, in[3] + 0x8f0ccc92, 10);
	MD5STEP(F4, c, d, a, b, in[10] + 0xffeff47d, 15);
	MD5STEP(F4, b, c, d, a, in[1] + 0x85845dd1, 21);
	MD5STEP(F4, a, b, c, d, in[8] + 0x6fa87e4f, 6);
	MD5STEP(F4, d, a, b, c, in[15] + 0xfe2ce6e0, 10);
	MD5STEP(F4, c, d, a, b, in[6] + 0xa3014314, 15);
	MD5STEP(F4, b, c, d, a, in[13] + 0x4e0811a1, 21);
	MD5STEP(F4, a, b, c, d, in[4] + 0xf7537e82, 6);
	MD5STEP(F4, d, a, b, c, in[11] + 0xbd3af235, 10);
	MD5STEP(F4, c, d, a, b, in[2] + 0x2ad7d2bb, 15);
	MD5STEP(F4, b, c, d, a, in[9] + 0xeb86d391, 21);

	buf[0] += a;
	buf[1] += b;
	buf[2] += c;
	buf[3] += d;
}

/*
 * Update context to reflect the concatenation of another buffer full
 * of bytes.
 */
static void
_MD5Update(_MD5_CTX *ctx, unsigned char const *buf, unsigned len)
{
	uint32_t t;

	/* Update bitcount */

	t = ctx->bits[0];
	if ((ctx->bits[0] = t + ((uint32_t) len << 3)) < t)
		ctx->bits[1]++;		/* Carry from low to high */
	ctx->bits[1] += len >> 29;

	t = (t >> 3) & 0x3f;	/* Bytes already in shsInfo->data */

	/* Handle any leading odd-sized chunks */

	if (t) {
		unsigned char *p = (unsigned char *) ctx->in + t;

		t = 64 - t;
		if (len < t) {
			memcpy(p, buf, len);
			return;
		}
		memcpy(p, buf, t);
		byteReverse(ctx->in, 16);
		_MD5Transform(ctx->buf, (uint32_t *) ctx->in);
		buf += t;
		len -= t;
	}
	/* Process data in 64-byte chunks */

	while (len >= 64) {
		memcpy(ctx->in, buf, 64);
		byteReverse(ctx->in, 16);
		_MD5Transform(ctx->buf, (uint32_t *) ctx->in);
		buf += 64;
		len -= 64;
	}

	/* Handle any remaining bytes of data. */

	memcpy(ctx->in, buf, len);
}

/*
 * Final wrapup - pad to 64-byte boundary with the bit pattern 
 * 1 0* (64-bit count of bits processed, MSB-first)
 */
static void
_MD5Final(unsigned char digest[16], _MD5_CTX *ctx)
{
	unsigned count;
	unsigned char *p;

	/* Compute number of bytes mod 64 */
	count = (ctx->bits[0] >> 3) & 0x3F;

	/* Set the first char of padding to 0x80.  This is safe since there is
	   always at least one byte free */
	p = ctx->in + count;
	*p++ = 0x80;

	/* Bytes of padding needed to make 64 bytes */
	count = 64 - 1 - count;

	/* Pad out to 56 mod 64 */
	if (count < 8) {
		/* Two lots of padding:  Pad the first block to 64 bytes */
		memset(p, 0, count);
		byteReverse(ctx->in, 16);
		_MD5Transform(ctx->buf, (uint32_t *) ctx->in);

		/* Now fill the next block with 56 bytes */
		memset(ctx->in, 0, 56);
	} else {
		/* Pad block to 56 bytes */
		memset(p, 0, count - 8);
	}
	byteReverse(ctx->in, 14);

	/* Append length in bits and transform */
	((uint32_t *) ctx->in)[14] = ctx->bits[0];
	((uint32_t *) ctx->in)[15] = ctx->bits[1];

	_MD5Transform(ctx->buf, (uint32_t *) ctx->in);
	byteReverse((unsigned char *) ctx->buf, 4);
	memcpy(digest, ctx->buf, 16);
	memset((char *) ctx, 0, sizeof(ctx));	/* In case it's sensitive */
}

/*
 * Stringify binary data. Output buffer must be twice as big as input,
 * because each byte takes 2 bytes in string representation
 */
static void
bin2str(char *to, const unsigned char *p, size_t len)
{
	const char	*hex = "0123456789abcdef";
	
	for (;len--; p++) {
		*to++ = hex[p[0] >> 4];
		*to++ = hex[p[0] & 0x0f];
		
	}
	*to = '\0';
}

/*
 * Return stringified MD5 hash for list of vectors.
 * buf must point to at least 32-bytes long buffer
 */
static void
md5(char *buf, ...)
{
	unsigned char		hash[16];
	const unsigned char	*p;
	va_list			ap;
	_MD5_CTX		ctx;

	_MD5Init(&ctx);

	va_start(ap, buf);
	while ((p = va_arg(ap, const unsigned char *)) != NULL)
		_MD5Update(&ctx, p, strlen((char *) p));
	va_end(ap);

	_MD5Final(hash, &ctx);
	bin2str(buf, hash, sizeof(hash));
}

/*
 * HTTP digest authentication
 */
struct digest {
	char		user[USER_MAX];
	char		uri[IO_MAX];
	char		cnonce[33];
	char		nonce[33];
	char		resp[33];
	char		qop[16];
	char		nc[16];
};

static void
fetchfield(const char **from, char *to, int len, int shift)
{
	const char	*p;

	*from += shift;

	if (**from == '"')
		(*from)++;
	for (p = *from; *p != '"' && p != '\0' && *p != ',' && *p != ' '; )
		p++;

	len--;	/* for '\0' terminator */
	if (p - *from < len)
		len = p - *from;
	(void) memcpy(to, *from, len);
	to[len] = '\0';
	*from += p - *from;
}

/*
 * Fetch a password provided by user.
 * Return 1 if success, 0 otherwise.
 */
static int
getauth(struct conn *c, struct digest *dig)
{
	const char	*p = c->auth, *e = p + strlen(c->auth);
	
	if (ncasecmp(p, "Digest ", 7) != 0)
		return (0);

	(void) memset(dig, 0, sizeof(dig));

	for (p += 7; p < e; p++)
		if (ncasecmp(p, "username=", 9) == 0)
			fetchfield(&p, dig->user, sizeof(dig->user), 9);
		else if (ncasecmp(p, "nonce=", 6) == 0)
			fetchfield(&p, dig->nonce, sizeof(dig->nonce), 6);
		else if (ncasecmp(p, "response=", 9) == 0)
			fetchfield(&p, dig->resp, sizeof(dig->resp), 9);
		else if (ncasecmp(p, "uri=", 4) == 0)
			fetchfield(&p, dig->uri, sizeof(dig->uri), 4);
		else if (ncasecmp(p, "qop=", 4) == 0)
			fetchfield(&p, dig->qop, sizeof(dig->qop), 4);
		else if (ncasecmp(p, "cnonce=", 7) == 0)
			fetchfield(&p, dig->cnonce, sizeof(dig->cnonce), 7);
		else if (ncasecmp(p, "nc=", 3) == 0)
			fetchfield(&p, dig->nc, sizeof(dig->cnonce), 3);

	elog(0, "[%s] [%s] [%s] [%s] [%s] [%s]",
	    dig->user, dig->uri, dig->resp, dig->qop, dig->cnonce, dig->nc);

	return (1);
}

/*
 * Check the user's password, return 1 if OK
 */
static int
checkpass(const struct conn *c, const char *a1, const struct digest *dig)
{
	char		a2[33], resp[33];

	/* nonce must be not too old, and uris must match */
	if (strcmp(dig->uri, c->ouri) != 0 ||
	    strlen(dig->resp) != 32 ||
	    now - strtoul(dig->nonce, NULL, 10) > 300)
		return (0);

	md5(a2, c->method, ":", c->ouri, NULL);
	md5(resp, a1, ":", dig->nonce, ":", dig->nc, ":",
	    dig->cnonce, ":", dig->qop, ":", a2, NULL);
	elog(0, "checkpass: [%s] [%s]", resp, dig->resp);

	return (strcmp(resp, dig->resp) == 0);
}

static FILE *
open_auth_file(const char *path)
{
	char 		name[FILENAME_MAX];
	const char	*p, *e;

	if (globhtpass) {
		/* Use global passwords file */
		(void) snprintf(name, sizeof(name), "%s", globhtpass);
	} else {
		/* Try to find .htpasswd in requested directory */
		for (p = path, e = p + strlen(p) - 1; e > p; e--)
			if (*e == '/')
				break;
		assert(*e == '/');
		(void) snprintf(name, sizeof(name), "%.*s/%s",
		    (int) (e - p), p, htpass);
	
		/* Fix directory separators */
		copypath(name, name, strlen(name) + 1);
	}

	return (fopen(name, "r"));
}

/*
 * Check authorization. Return 1 if not needed or authorized, 0 otherwise
 */
static int
checkauth(struct conn *c, const char *file)
{
	int		authorized = 1;
	char		line[512], user[65], ha1[65], realm[65];
	struct digest	digest;
	FILE		*fp = NULL;

#ifdef EMBEDDED
	struct userauth	*uap;
	struct llhead	*lp;

	/* Check, is this URL protected by shttpd_protect_url() */
	LL_FOREACH(&auths, lp) {
		uap = LL_ENTRY(lp, struct userauth, link);
		elog(0, "checkauth: %s %s %s", c->ouri,uap->url, uap->filename);
		if (strncmp(uap->url, c->ouri, strlen(uap->url)) == 0) {
			fp = fopen(uap->filename, "r");
			break;
		}
	}
#endif /* EMBEDDED */

	if (fp == NULL)
		fp = open_auth_file(file);

	/* Try to open passwords file. If opened, set authorized to 0  */
	if (fp != NULL) {
		authorized = 0;

		if (c->auth && getauth(c, &digest) &&
		    (c->user = mystrdup(digest.user)) != NULL)
			/* Each line should be "USER:REALM:HASH" */
			while (fgets(line, sizeof(line), fp) != NULL) {
				if (sscanf(line, "%64[^:]:%64[^:]:%64s",
				    user, realm, ha1) != 3)
					continue;
				elog(0, "checkauth: [%s] [%s,%s] [%s,%s]",
				    line, c->user, user, domain, realm);
				if (!strcmp(c->user, user) &&
				    !strcmp(domain, realm)) {
					authorized = checkpass(c, ha1, &digest);
					break;
				}
			}
		(void) fclose(fp);
	}

	return (authorized);
}
#endif /* NO_AUTH */

#ifndef NO_CGI
/*
 * Open program, and return read/write pipes to it
 * Return 0 if OK, -1 if error
 */
static int
redirect(struct conn *c, char *prog, char *envblk, char **envp)
{
	int			pair[2];
	char			dir[FILENAME_MAX], *p;
#ifdef _WIN32
	int			sock;	/* non-inheritable copy */
	char			cmdline[FILENAME_MAX], line[FILENAME_MAX];
	FILE			*fp;
	STARTUPINFO		si;
	PROCESS_INFORMATION	pi;
#else
	pid_t			pid;
#endif /* _WIN32 */
	
	elog(0, "%s exit", __func__);
	if (mysocketpair(pair) != 0)
		return (-1);

	/* CGI must be executed in its own directory */
	(void) strcpy(dir, prog);
	for (p = dir + strlen(dir) - 1; p > dir; p--)
		if (*p == '/') {
			*p++ = '\0';
			break;
		}
	copypath(dir, dir, strlen(dir) + 1);

#ifdef _WIN32
	if (!DuplicateHandle(GetCurrentProcess(),(HANDLE) pair[0],
	    GetCurrentProcess(), (HANDLE *) &sock,
	    0, FALSE, DUPLICATE_SAME_ACCESS))
		elog(1, "main: DuplicateHandle: %d", ERRNO);
	(void) closesocket(pair[0]);
	(void) memset(&si, 0, sizeof(si));
	(void) memset(&pi, 0, sizeof(pi));

	si.cb = sizeof(si);
	si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
	si.wShowWindow = SW_HIDE;
	si.hStdInput = si.hStdOutput = si.hStdError = (HANDLE) pair[1];

	/* If CGI file is a script, try to read the interpreter line */
	if ((fp = fopen(prog, "r")) != NULL) {
		(void) fgets(line, sizeof(line), fp);
		if (memcmp(line, "#!", 2) != 0)
			line[2] = '\0';
		/* Trim whitespaces from interpreter name */
		for (p = &line[strlen(line) - 1]; p > line && isspace(*p); p--)
			*p = '\0';
		(void) fclose(fp);
	}

	/* Prepare command line */
	(void) snprintf(cmdline, sizeof(cmdline), "%s%s%s",
	    line + 2, line[2] == '\0' ? "" : " ", prog);
#if 0
	copypath(cmdline, cmdline, strlen(cmdline) + 1);
#endif
	elog(0, "redirect: cmdline [%s], dir [%s]", cmdline, dir);

	if (CreateProcess(NULL, cmdline, NULL, NULL, TRUE,
	    0, envblk, dir, &si, &pi) == 0) {
		elog(0, "redirect: CreateProcess(%s): %d", cmdline, ERRNO);
		(void) closesocket(sock);
		(void) closesocket(pair[1]);
		return (-1);
	} else {
#if 0
		CloseHandle(pi.hProcess);
		(void) closesocket(pair[1]);
#endif
		c->hChild = pi.hProcess;
		c->childsock = pair[1];
		CloseHandle(pi.hThread);
		c->cgisock = pair[0];
	}
	return (sock);
#else
	envblk = NULL;	/* unused */

	if ((pid = fork()) == -1) {
		elog(0, "redirect: fork");
		(void) closesocket(pair[0]);
		(void) closesocket(pair[1]);
		return (-1);
	} else if (pid == 0) {
		/* Child */
		(void) closesocket(pair[0]);
		(void) dup2(pair[1], 0);
		(void) dup2(pair[1], 1);
		(void) dup2(pair[1], 2);
		(void) chdir(dir);
		(void) execle(prog, prog, NULL, envp);
		elog(1, "redirect: exec(%s)", prog);	/* Error, exit*/
	} else {
		/* Parent */
		(void) closesocket(pair[1]);
		c->cgisock = pair[0];
	}
	elog(0, "%s exit", __func__);
	
	return (pair[0]);
#endif /* _WIN32 */
}

/*
 * Prepare the environment for the CGI program, and start CGI program.
 */
static int
spawncgi(struct conn *c, char *prog)
{
	int	i, n = 0, k = 0, ret;
	char	path[FILENAME_MAX], env[ARG_MAX], *penv[64], *p;
	char	*s, var[IO_MAX], val[IO_MAX];

#define	ADDENV(x)		\
	penv[k++] = &env[n], \
	n += 1 + snprintf(env + n, sizeof(env) - n -1, x), \
	snprintf_sanitize(sizeof(env), &n)
#define	ADDENV1(x, y)		if (y) \
	penv[k++] = &env[n], \
	n += 1 + snprintf(env + n, sizeof(env)-n-1, x, y), \
	snprintf_sanitize(sizeof(env), &n)
#define	ADD_IF_SET(x)	if ((p = getenv(x)) != NULL) ADDENV1(x "=%s", p);
	
	elog(0, "%s enter [%s]", __func__, prog);
	
	/* Prepare the environment block */
	ADDENV("GATEWAY_INTERFACE=CGI/1.1");
	ADDENV("SERVER_PROTOCOL=HTTP/1.1");
	ADDENV1("SERVER_PORT=%u", lport);
	ADDENV1("SERVER_NAME=%s", domain);
	ADDENV1("DOCUMENT_ROOT=%s", docroot);
	ADDENV1("REQUEST_METHOD=%s", c->method);
	ADDENV("REDIRECT_STATUS=200");			/* For PHP */
	p = c->query_string ? c->query_string : "";
	ADDENV1("QUERY_STRING=%s", p);
	ADDENV1("REMOTE_ADDR=%s", inet_ntoa(c->sa.u.sin.sin_addr));

	ADDENV1("SCRIPT_NAME=%s", c->uri);
	ADDENV1("CONTENT_TYPE=%s", c->ctype);
	ADD_IF_SET("PATH");
	ADD_IF_SET("COMSPEC");
	ADD_IF_SET("LD_LIBRARY_PATH");
	ADD_IF_SET("PERLLIB");
	
	/* Set up PAHT_INFO and PATH_TRANSLATED */
	if (c->pathinfo != NULL) {
		(void) snprintf(path, sizeof(path), "/%s", c->pathinfo);
		ADDENV1("PATH_INFO=%s", path);
	}
	ADDENV1("PATH_TRANSLATED=%s", prog);
	
	if (c->user) {
		ADDENV1("REMOTE_USER=%s", c->user);
		ADDENV("AUTH_TYPE=Digest");
	}

	if (c->clength > 0)
		ADDENV1("CONTENT_LENGTH=%lu", c->clength);

	/* Add all headers as HTTP_* variables */
	for (s = strchr(c->remote.buf, '\n') + 1;
	    k < (int) NELEMS(penv) - 2 && n < (int) sizeof(env) - 2 &&
	    s != NULL && s[0] != '\n' && s[0] != '\r'; s = p) {
		
		if ((p = strchr(s, '\n')) != NULL)
			p++;
		
		if (sscanf(s, "%[-a-zA-Z0-9]: %[^\r\n]", var, val) != 2)
			continue;
		
		for (i = 0; i < (int) strlen(var); i++) {
			var[i] = toupper(((unsigned char *) var)[i]);
			if (var[i] == '-')
				var[i] = '_';
		}
		penv[k++] = &env[n];
		n += 1 + snprintf(env + n, sizeof(env) - n - 1,
		    "HTTP_%s=%s", var, val);
		snprintf_sanitize(sizeof(env), &n);
	}

	penv[k++] = NULL;
	env[n++] = '\0';

	assert(k < (int) NELEMS(penv));
	assert(n < (int) NELEMS(env));
	ret  = redirect(c, prog, env, penv);
	elog(0, "%s exit %d", __func__, ret);
	
	return (ret);
}

static void
cgiparse(struct conn *c)
{
	char	l[64];
	int	n;

	elog(0, "%s enter", __func__);
	if ((n = getreqlen(c->local.buf, c->local.nread)) == REQ_INVALID) {
		senderr(c, 500, "Error", "", "Script sent invalid headers");
	} else if (n == REQ_INCOMPLETE) {
		c->local.nwritten = c->local.nread;
		elog(0, "%s [%.*s]", __func__, c->local.nread, c->local.buf);
	} else {
				
		/* Received all headers. Set status. */
		parseheaders(c, c->local.buf);
		if (c->location)
			c->status = 302;
		if (c->status == 0)
			c->status = 200;

		/* Output the status line */
		n = snprintf(l, sizeof(l),  "HTTP/1.1 %u OK\r\n", c->status);
		snprintf_sanitize(sizeof(l), &n);
		(void) writeremote(c, l, n);

		/* Set flags, so script output will be passed directly out */
		c->cgiparsed++;
		c->local.nwritten = 0;
	}
	elog(0, "%s exit", __func__);
}

/*
 * GET or POST for cgi scripts
 */
static void
get_cgi(struct conn *c)
{
	int	len, n;

	elog(0, "%s enter", __func__);
	assert(c->cgisock != -1);

	/* Push data received from remote side to a CGI program */
	if (c->http_method == HTTP_POST && (len = IO_DATALEN(&c->remote)) > 0) {
		n = send(c->cgisock, c->remote.buf + c->remote.nwritten,len, 0);
		elog(0, "get_cgi: sent %d out of %d [%d]", n, len, ERRNO);
		if (n > 0) {
			io_inc_nwritten(&c->remote, n);
#if 0
		} else if (n == 0 || (n == -1 && ERRNO == EWOULDBLOCK)) {
			c->local.done++;
#endif
		}
	}

#if _WIN32
	/*
	 * On windows, if child CGI process terminates, recv() from cgisock
	 * return -1 with errno ECONRESET. There is no way to receive
	 * the rest of data from socket, so we loose some data if we're
	 * not reading fast enough. So we do not close our part of the
	 * socket pair, and save the child process handle.
	 * We do not close pair[0] so we never get ECONRESET.
	 * we use handle to check is the child still running. If it has
	 * been terminated, we close everything.
	 */
	if (c->local.done)
		TerminateProcess(c->hChild, 0);
	if (WaitForSingleObject(c->hChild, 1) == 0) {
		(void) CloseHandle(c->hChild);
		(void) closesocket(c->childsock);
	}
#endif /* _WIN32 */

	
	/*
	 * Script may output Status: and Location: headers,
	 * which may alter the response code. So buffer in headers,
	 * parse them, send correct status code and then forward
	 * all data from CGI script back to the remote end.
	 */
	len = IO_SPACELEN(&c->local);
	if ((n = recv(c->cgisock, c->local.buf + c->local.nread, len, 0)) > 0) {
		c->local.nread += n;
		if (c->cgiparsed == 0)
			cgiparse(c);
	} else if (n == 0 || (n == -1 && ERRNO != EWOULDBLOCK)) {
		c->local.done++;
		c->io = NULL;
		if (c->cgiparsed == 0)
			senderr(c, 500, "Error", "","Bad headers sent: [%.*s]",
			c->local.nread, c->local.buf);
	}

	elog(0, "%s exit, recv-ed %d out of %d [%d]", __func__, n, len, ERRNO);
}

/*
 * GET or POST cgi file
 */
static void
do_cgi(struct conn *c)
{
	elog(0, "%s enter", __func__);
	accesslog(c, 200);
	c->remote.nwritten = c->reqlen;
	io_inc_nwritten(&c->remote, 0);
	(void) nonblock(c->cgisock);
	if (c->http_method == HTTP_GET)
		c->remote.nread = 0;
	c->io = get_cgi;
	elog(0, "%s exit", __func__);
}

/*
 * Verify that given file has CGI extension
 */
static int
iscgi(const char *path)
{
	int	len1 = strlen(path), len2 = strlen(cgipat);
	return  (len1 > len2 && !ncasecmp(path + len1 - len2, cgipat, len2));
}
#endif /* NO_CGI */

/*
 * For given directory path, substitute it to valid index file.
 * Return 0 if index file has been found, -1 if not found
 */
static int
useindex(char *path, size_t maxpath)
{
	struct stat	st;
	char		try[FILENAME_MAX], name[FILENAME_MAX];
	const char	*s = ifile, *p;
	int		ret = -1;
	size_t		len;

	elog(0, "%s enter [%s]", __func__, path);
	do {
		if ((p = strchr(s, ',')) != NULL) {
			len = p - s;
			assert(len < sizeof(name));
			(void) memcpy(name, s, len);
			name[len] = '\0';
		} else {
			(void) strncpy(name, s, sizeof(name));
			name[sizeof(name) - 1] = '\0';
		}

		(void) snprintf(try, sizeof(try), "%s%s", path, name);
		elog(0, "%s: trying [%s]", __func__, try);
		if (mystat(try, &st) == 0) {
			(void) strncpy(path, try, maxpath);
			path[maxpath - 1] = '\0';
			ret = 0;
		}

		/* Move to the next index file */
		s = p ? p + 1 : NULL;
	} while (ret == -1 && s != NULL);
	elog(0, "%s exit [%d]", __func__, ret);

	return (ret);
}

/*
 * Handle request
 */
static void
decide(struct conn *c)
{
	char		path[FILENAME_MAX];
	size_t		docrootlen;
#ifdef EMBEDDED
	struct userurl	*p;
#endif /* EMBEDDED */

	elog(0, "%s enter, uri [%s]", __func__, c->uri);
	
	urldecode(c->uri, c->uri);
	if ((c->query_string = strchr(c->uri, '?')) != NULL) {
		*c->query_string++ = '\0';
		c->query_string_len = strlen(c->query_string);
	}
	docrootlen = strlen(docroot);
	(void) snprintf(path, sizeof(path), "%s%s", docroot, c->uri);
	killdots(path + docrootlen);

#ifndef NO_AUTH
	if (checkauth(c, path) != 1) {
		char	buf[256];
		(void) snprintf(buf, sizeof(buf),
		    "WWW-Authenticate: Digest qop=\"auth\", realm=\"%s\", "
		    "nonce=\"%lu\"", domain, (unsigned long) now);

		senderr(c, 401, "Unauthorized", buf, "Authorization required");
	} else
#endif /* NO_AUTH */

	/* If directory has been requested, substitute index file */	
#if EMBEDDED
	if ((p = isregistered(c->uri)) != NULL) {
		do_embedded(c, p);
	} else
#endif /* EMBEDDED*/
	
	if (path[strlen(path) - 1] == '/' && list_dirs == 0 &&
	    (c->http_method == HTTP_GET || c->http_method == HTTP_POST) &&
	    useindex(path, sizeof(path) - 1) == -1) {
		senderr(c, 403, "Forbidden", "", "Directory Listing Denied");
	} else if (path[strlen(path) - 1] == '/' && list_dirs == 1 &&
	    (c->http_method == HTTP_GET || c->http_method == HTTP_POST) &&
	    useindex(path, sizeof(path) - 1) == -1) {
		if ((c->path = mystrdup(path)))
			do_dir(c);
		else
			senderr(c, 500, "Error", "", "strdup");
	} else if (strstr(path, htpass)) {
		senderr(c, 403, "Forbidden","", "Permission Denied");
#ifdef WITH_PUT_AND_DELETE
	} else if (c->http_method == HTTP_PUT) {
		int	rc;
		if ((rc = put_dir(path)) == 0) {
			senderr(c, 200, "OK","","");
		} else if (rc == -1) {
			senderr(c, 500, "Error","","%s", strerror(errno));
		} else if ((c->wfd = myopen(path, O_WRONLY | O_BINARY |
		    O_CREAT | O_NONBLOCK | O_TRUNC, 0644)) == -1) {
			elog(0, "handle: open(%s): %s", path, strerror(errno));
			senderr(c, 500, "Error","","PUT error");
		} else {
			accesslog(c, 200);
			c->local.nread = snprintf(c->local.buf,
			    sizeof(c->local.buf), "%s",
			    "HTTP/1.1 200 OK\r\n\r\n");
			snprintf_sanitize(sizeof(c->local.buf),
			    (int *) &c->local.nread);
			c->remote.nwritten = c->reqlen;
			io_inc_nwritten(&c->remote, 0);
			(void) nonblock(c->wfd);
			c->io = put_file;
		}
	} else if (c->http_method == HTTP_DELETE) {
		if (remove(path) == 0)
			senderr(c, 200, "OK", "", "");
		else
			senderr(c, 500, "Error", "", "%s", strerror(errno));
#endif /* WITH_PUT_AND_DELETE */
	} else if (getpathinfo(path, &c->st, &c->pathinfo) != 0 ||
	    (((c->st.st_mode & S_IFDIR) && c->pathinfo != NULL))) {
		elog(0, "decide: stat(%s): %d", path, ERRNO);
		senderr(c, 404, "Not Found","", "Not Found");
	} else if ((c->st.st_mode & S_IFDIR) && list_dirs) {
		if ((c->path = mystrdup(path)))
			do_dir(c);
		else
			senderr(c, 500, "Error", "", "strdup");
	} else if ((c->st.st_mode & S_IFDIR) && list_dirs == 0) {
		elog(0, "decide: %s: Denied", path);
		senderr(c, 403, "Forbidden","",
		    "Directory listing denied");
#ifndef NO_CGI
	} else if (iscgi(path)) {
		if ((c->cgisock = spawncgi(c, path)) == -1) {
			senderr(c, 500, "Server Error", "",
			    "Error executing CGI script");
		} else {
			do_cgi(c);
		}
#endif /* NO_CGI */
	} else if (c->ims && c->ims >= c->st.st_mtime) {
		elog(0, "decide: %s: Not Modified", path);
		senderr(c, 304, "Not Modified","", "");
	} else if ((c->rfd = myopen(path, O_RDONLY | O_BINARY, 0644)) != -1) {
		accesslog(c, 200);
		do_get(c);
	} else {
		senderr(c, 500, "Error", "", "Internal Error");
		c->done++;
	}
	elog(0, "%s exit", __func__);
}

/*
 * Parse HTTP request
 */
static void
parse(struct conn *c, const char *s)
{
	/* Will pre-process in something like "15s 1024s 15s" */
	const char	*fmt =
		"%" _STR2(METHOD_MAX) "s "
		"%" _STR2(IO_MAX) "s "
		"%" _STR2(METHOD_MAX) "s";

	elog(0, "%s enter [%s] [%s]", __func__, fmt, s);
	
	/* Get the request line */
	if (sscanf(s, fmt, c->method, c->uri, c->proto) != 3) {
		senderr(c, 404, "Bad Request","", "Bad Request");
	} else if (c->uri[0] != '/') {
		senderr(c, 404, "Bad Request","", "Bad Request");
	} else if (ncasecmp(c->proto, "HTTP", 4) != 0) {
		senderr(c, 501, "Bad Protocol","", "Procotol Not Supported");
	} else if (ncasecmp(c->method, "GET", 3) == 0) {
		c->http_method = HTTP_GET;
	} else if (ncasecmp(c->method, "POST", 4) == 0) {
			c->http_method = HTTP_POST;
#ifdef WITH_PUT_AND_DELETE
	} else if (ncasecmp(c->method, "PUT", 3) == 0) {
		c->http_method = HTTP_PUT;
	} else if (ncasecmp(c->method, "DELETE", 6) == 0) {
		c->http_method = HTTP_DELETE;
#endif /* WITH_PUT_AND_DELETE */
	} else {
		senderr(c, 501, "Not Implemented", "","Method Not Implemented");
	}
	
	/* XXX senderr() should set c->local.done flag! */
	if (c->local.done == 0) {
		parseheaders(c, strchr(s, '\n') + 1);
		(void) strcpy(c->ouri, c->uri);
		decide(c);
		c->parsed++;
	}
	elog(0, "%s exit", __func__);
}

/*
 * Add mime type to the known mime type list
 */
static void
addmimetype(const char *ext, const char *mime)
{
	struct mimetype	*p;

	/* XXX possible resource leak  */
	if ((p = calloc(1, sizeof(*p))) != NULL &&
	    (p->ext = mystrdup(ext)) != NULL &&
	    (p->mime = mystrdup(mime)) != NULL) {
		p->extlen = strlen(p->ext);
		p->next = mimetypes;
		mimetypes = p;
	}
}

/*
 * Free mime types list
 */
static void
freemime(void)
{
	struct mimetype	*p, *tmp;

	for (p = mimetypes; p != NULL; p = tmp) {
		tmp = p->next;
		free(p->ext);
		free(p->mime);
		free(p);
	}
}

/*
 * Set mime type list from file
 */
static void
setmimetypes(const char *file)
{
	FILE	*fp;
	char	line[512], ext[sizeof(line)], mime[sizeof(line)];
	
	if ((fp = fopen(file, "r")) == NULL)
		elog(1, "setmimetypes: fopen(%s): %s", file, strerror(errno));

	while (fgets(line, sizeof(line), fp) != NULL)
		if (sscanf(line, "%s %s", ext, mime) == 2)
			addmimetype(ext, mime);

	(void) fclose(fp);
}

/*
 * Set default mime types list
 */
static void
setdefmimetypes(void)
{
	addmimetype("jpeg", "image/jpeg");
	addmimetype("png", "image/png");
	addmimetype("mpg", "video/mpeg");
	addmimetype("asf", "video/x-ms-asf");
	addmimetype("avi", "video/x-msvideo");
	addmimetype("bmp", "image/bmp");
	addmimetype("doc", "application/msword");
	addmimetype("exe", "application/octet-stream");
	addmimetype("zip", "application/zip");
	addmimetype("jpg", "image/jpeg");
	addmimetype("gif", "image/gif");
	addmimetype("txt", "text/plain");
	addmimetype("c", "text/plain");
	addmimetype("h", "text/plain");
	addmimetype("ico", "image/x-icon");
	addmimetype("css", "text/css");
	addmimetype("htm", "text/html");
	addmimetype("html", "text/html");
}

/*
 * Process given connection
 */
static void
serve(struct conn *c, fd_set *rset, fd_set *wset)
{
	int	n, len;

	elog(0, "%s enter %p: local %d.%d.%d, remote %d.%d.%d", __func__, c,
	    c->local.done, c->local.nread, c->local.nwritten,
	    c->remote.done, c->remote.nread, c->remote.nwritten);

	/* Read from remote end  */
	assert(c->sock != -1);
	if ((len = IO_SPACELEN(&c->remote) - 1)> 0 && FD_ISSET(c->sock, rset)) {
		n = readremote(c, c->remote.buf + c->remote.nread, len);
		if (n > 0) {
			c->remote.nread += n;
			c->remote.buf[c->remote.nread] = '\0';
			c->expire += EXPIRE_TIME;
		} else if (c->parsed == 0 && c->remote.done) {
			c->done++;
		}
	}

	if (c->parsed == 0 && IO_SPACELEN(&c->remote) <= 1)
		senderr(c, 400, "Bad Request", "", "Bad request");
	else if (c->parsed == 0) {
		c->reqlen = getreqlen(c->remote.buf, c->remote.nread);
		if (c->reqlen == REQ_INVALID)
			senderr(c, 400, "Bad Request", "", "Bad request");
		else if (c->reqlen > 0)
			parse(c, c->remote.buf);
	}

	if (c->io && IO_SPACELEN(&c->local)) {
		c->io(c);
		c->expire += EXPIRE_TIME;
	}

	if ((len = IO_DATALEN(&c->local)) > 0 && FD_ISSET(c->sock, wset) &&
	    (n = writeremote(c, c->local.buf + c->local.nwritten, len)) > 0)
		io_inc_nwritten(&c->local, n);

	if ((c->remote.done && c->remote.nread == 0) ||
	    (c->local.done && c->local.nread == 0))
		c->done++;
	
	elog(0, "%s exit %p: local %d.%d.%d, remote %d.%d.%d", __func__, c,
	    c->local.done, c->local.nread, c->local.nwritten,
	    c->remote.done, c->remote.nread, c->remote.nwritten);
}

/*
 * Add file descriptors of opened connections to the given
 * file descriptors set. Bump up the maxfd.
 */
void
shttpd_merge_fds(fd_set *rfds, fd_set *wfds, int *maxfd)
{
	struct llhead	*lp;
	struct conn	*c;
	int		n = 0;

#define	MERGE_FD(fd,set) \
	do { FD_SET(fd,set); if (fd > *maxfd) *maxfd = fd; } while (0)
	
	LL_FOREACH(&conns, lp) {
		c = LL_ENTRY(lp, struct conn, link);
		n++;	/* Number of active connections */
		
		MERGE_FD(c->sock, rfds); /* Remote socket always in read set */
		
		if (IO_DATALEN(&c->local))
			MERGE_FD(c->sock, wfds);
		
		if (c->cgisock != -1)
			MERGE_FD(c->cgisock, rfds);
		else if (c->dummy[0] != -1)
			MERGE_FD(c->dummy[0], rfds);
		else if (c->wfd != -1 && IO_DATALEN(&c->remote))
			MERGE_FD(c->wfd, wfds);
		else if (c->rfd != -1)
			MERGE_FD(c->rfd, rfds);
	}

	if (n > nconns)
		nconns = n;	/* Maximum active connections handled */
}

/*
 * One iteration of server loop.
 */
void
shttpd_poll(int listen_socket, unsigned milliseconds)
{
	struct llhead	*lp, *tmp;
	struct usa	sa;
	struct conn	*c;
	int		sock, maxfd = listen_socket;
	struct timeval	tv;			/* Timeout for select() */
	fd_set		rset, wset;

	now = time(0);
	sa.len = sizeof(sa.u.sin);
	
	/* Populate the descriptor set */
	FD_ZERO(&rset);
	FD_ZERO(&wset);
	if (listen_socket != -1)
		FD_SET(listen_socket, &rset);

	shttpd_merge_fds(&rset, &wset, &maxfd);
	tv.tv_sec = milliseconds / 1000;
	tv.tv_usec = milliseconds % 1000;

	/* Find out who is ready for IO */
	if (select(maxfd + 1, &rset, &wset, NULL, &tv) < 0) {
		elog(0, "select: %s", strerror(errno));
	} else {
		/* If listening socket is ready, accept new connection */
		if (listen_socket != -1 && FD_ISSET(listen_socket, &rset) &&
		    (sock = accept(listen_socket, &sa.u.sa, &sa.len)) != -1)
			newconnection(sock, &sa);

		/* Loop through all connections, handle if needed */
		LL_FOREACH_SAFE(&conns, lp, tmp) {
			c = LL_ENTRY(lp, struct conn, link);
			serve(c, &rset, &wset);
			if (c->done || c->expire < now)
				disconnect(c);
		}
	}
}

/*
 * Initialize shttpd
 */
void
shttpd_init(void)
{
	/* Init global variable 'now', needed for elog() etc */
	now = time(0);

	/* Initialize document root if not set from cmd line */
	if (docroot[0] == '\0')
		(void) getcwd(docroot, sizeof(docroot));

	/* Redirect stderr to a file if requested */
	if (errfile)
		errtofile(errfile);
	
	/* Set mime types */
	if (mimefile)
		setmimetypes(mimefile);
	else
		setdefmimetypes();

	/* Open access log file */
	if ((logfp = fopen(logfile, "a")) == NULL)
		elog(0, "cannot open logfile %s: %s",
		    logfile, strerror(errno));

#ifdef _WIN32
	/* Initialize winsock library */
	{
		WSADATA data;
		int	opt = SO_SYNCHRONOUS_NONALERT;
		WSAStartup(MAKEWORD(2,2), &data);
		if (setsockopt(INVALID_SOCKET, SOL_SOCKET, SO_OPENTYPE,
		    (char *)&opt,sizeof(opt)) < 0)
			elog(1, "setsockopt: %d", ERRNO);
	}
#endif /* _WIN32 */

#if WITH_SSL
	if (dossl) {
		/* Initialize SSL crap */
		SSL_load_error_strings();
		SSLeay_add_ssl_algorithms();
		if ((ctx = SSL_CTX_new(SSLv23_server_method())) == NULL)
			elog(1, "main: error: SSL_CTX_new %p", ctx);
		else if (SSL_CTX_use_certificate_file(ctx, certfile,
		    SSL_FILETYPE_PEM) == 0)
			elog(1, "main: please put %s here", certfile);
		else if (SSL_CTX_use_PrivateKey_file(ctx, certfile,
		    SSL_FILETYPE_PEM) == 0)
			elog(1, "main: please put %s here", certfile);
	}
#endif /* WITH_SSL */
}

/*
 * Deallocate shttpd object, free up the resources
 */
void
shttpd_fini(void)
{

	/* Free allocated mime-types */
	freemime();

	/* If log stream was opened, close it now */
	if (logfp != NULL)
		(void) fclose(logfp);
}

#ifdef EMBEDDED
void
shttpd_register_url(const char *url, shttpd_callback_t cb, void *data)
{
	struct userurl	*p;

	if ((p = calloc(1, sizeof(*p))) != NULL) {
		p->func = cb;
		p->data = data;
		p->url = mystrdup(url);
		LL_TAIL(&urls, &p->link);
	}
}

void
shttpd_protect_url(const char *url, const char *filename)
{
	struct userauth	*p;

	if ((p = calloc(1, sizeof(*p))) != NULL) {
		p->url = mystrdup(url);
		p->filename = mystrdup(filename);
		LL_TAIL(&auths, &p->link);
	}
}

const char *
shttpd_get_var(struct conn *conn, const char *var)
{
	char	*p, *val, *e;
	int	len = strlen(var);

	for (p = conn->query_string, e = p + conn->query_string_len - len,
	    val = NULL; p != NULL && p < e; p++) {
		if (memcmp(p, var, len) == 0 && p[len] == '=')
			val = p += len + 1;
		while (*p != '\0' && *p != '&')
			p++;
		*p = '\0';
	}
	
	return (val);
}

void
shttpd_set_options(const char *root, int do_ssl, int do_debug)
{
	(void) strncpy(docroot, root, sizeof(docroot));
	docroot[sizeof(docroot) - 1] = '\0';
	debug	= do_debug;
	dossl	= do_ssl;
}

static void
substitute(char *buf, const char *kw, const char *subst)
{
	char	buf2[IO_MAX], *s = buf, *p, *p2 = buf2;
	int	len, kwlen = strlen(kw), slen = strlen(subst);

	do {
		if ((p = strstr(s, kw)) == NULL) {
			len = strlen(s);
			(void) memcpy(p2, s, len);
			p2 += len;
			s += len;
		} else {
			len = p - s;
			
			if (&p2[len] < &buf2[sizeof(buf2)]) {
				(void) memcpy(p2, s, len);
				p2 += len;
			}

			if (&p2[slen] < &buf2[sizeof(buf2)]) {
				(void) memcpy(p2, subst, slen);
				p2 += slen;
			}

			s = p + kwlen;
		}

	} while (*s != '\0');

	assert(p2 < &buf2[sizeof(buf2)]);
	*p2 = '\0';

	(void) memcpy(buf, buf2, p2 - buf2 + 1);
}

int
shttpd_template(struct conn *conn, const char *headers, const char *file, ...)
{
	va_list		ap;
	char		buf[IO_MAX], *keyword, *subst;
	FILE		*fp;
	int		n = 0;

	/* Read the file in */
	if ((fp = fopen(file, "r")) == NULL) {
		elog(0, "%s: fopen(%s): %s", __func__, file, strerror(errno));
	} else if ((n = fread(buf, 1, sizeof(buf) - 1, fp)) <= 0) {
		elog(0, "%s: fread(%s): %s", __func__, file, strerror(errno));
		(void) fclose(fp);
	} else {
		buf[n] = '\0';
		
		va_start(ap, file);
		while ((keyword = va_arg(ap, char *)) != NULL &&
		    (subst = va_arg(ap, char *)) != NULL)
			if (keyword[0] != '\0')
				substitute(buf, keyword, subst);
		va_end(ap);

		n = snprintf(conn->local.buf, sizeof(conn->local.buf),
		    "%s%s\r\n%s", "HTTP/1.1 200 OK\r\n", headers, buf);
		snprintf_sanitize(sizeof(conn->local.buf), &n);
		
		(void) fclose(fp);
	}

	return (n);
}

#ifdef MT
int
shttpd_push(struct conn *conn, const void *buf, size_t len)
{
	int		n;
	size_t		left = len;
	const char	*p = buf;

	do {
		n = writeremote(conn, p, left);
		if (n <= 0)
			return (-1);
		left -= n;
		p += n;
	} while (left > 0);

	return (len);
}

int
shttpd_printf(struct conn *conn, const char *fmt, ...)
{
	va_list	ap;
	int	len;
	char	buf[IO_MAX];

	va_start(ap, fmt);
	len = vsnprintf(buf, sizeof(buf), fmt, ap);
	snprintf_sanitize(sizeof(buf), &len);
	va_end(ap);

	return (shttpd_push(conn, buf, len));
}
#endif /* MT */

#else

/*
 * Stand-alone configuration. Define signal handlers, usage() and main()
 */

/*
 * SIGTERM, SIGINT signal handler
 */
static void
sigterm(int signo)
{
	elog(0, "%s enter", __func__);
	quit = signo;
	elog(0, "%s exit", __func__);
}

/*
 * Grim reaper of innocent children: SIGCHLD signal handler
 */
static void
sigchild(int signo)
{
	elog(0, "%s enter", __func__);
	while (waitpid(-1, &signo, WNOHANG) > 0) ;
	elog(0, "%s exit", __func__);
}


/*
 * Show usage string and exit.
 */
static void
usage(const char *prog)
{
	(void) fprintf(stderr,
	    "shttpd version %s (c) Sergey Lyubka\n"
	    "usage: %s [OPTIONS]\n"
	    "-d directory	wwwroot (dflt: %s)\n"
	    "-D 		no directory listing\n"
	    "-p port		listening port (dflt: 80 or 443)\n"
	    "-l logfile 	log file (dflt: %s)\n"
	    "-I 		inetd mode (dflt: no)\n"
	    "-e elogfile 	error log (dflt: none)\n"
	    "-i file1,... 	index files (dflt: %s)\n"
	    ,VERSION, prog, docroot, logfile, ifile);

	(void) fprintf(stderr,
#ifndef NO_CGI
	    "-c cgipattern	CGI file pattern (dflt: %s)\n"
#endif /* NO_CGI */
#ifndef NO_AUTH
	    "-P passfile	global auth file (dflt: none)\n"
#endif /* NO_AUTH */
	    "-u login_name	switch UID to user (dflt: none)\n"
	    "-N servername 	server domain name (auth realm) (dflt: %s)\n"
	    "-m mime-file 	mime types file (dflt: none)\n"
	    "-v 		debug mode\n"
#ifdef WITH_SSL
	    "-s certfile	SSL certfileificate (dflt: %s)\n"
	    "-S 		use SSL (dflt: no)\n"
#endif /* WITH_SSL */
#ifndef NO_CGI
	    , cgipat
#endif /* NO_CGI */
	    , domain
#ifdef WITH_SSL
	    , certfile
#endif /* WITH_SSL */
	    );

	exit(EXIT_FAILURE);
}

int
main(int argc, char *argv[])
{
	int		i, lsn;
	const char	*optarg;

	/*
	 * Parse command-line options. Not every system has getopt()
	 * function, so we use the good old method.
	 */
#define	GETARG(p)							\
	if ((p = (argv[i][2] ? &argv[i][2] : argv[++i])) == NULL)	\
		usage(argv[0])

	for (i = 1; i < argc && argv[i][0] == '-'; i++)
		switch (argv[i][1]) {

		case 'd':		/* Document root */
			GETARG(optarg);
			(void) strncpy(docroot, optarg, sizeof(docroot));
			docroot[sizeof(docroot) - 1] = '\0'; /* paranoia */
			break;

		case 'p':		/* shttpd_open_porting port */
			GETARG(optarg);
			lport = (uint16_t) atoi(optarg);
			break;

		case 'l':		/* Log file */
			GETARG(optarg);
			logfile = optarg;
			break;

		case 'i':		/* Index file */
			GETARG(optarg);
			ifile = optarg;
			break;

		case 'I':		/* Inetd mode */
			inetd++;
			break;

		case 'P':		/* Passwords file */
			GETARG(optarg);
			globhtpass = optarg;
			break;

		case 'c':		/* CGI pattern */
			GETARG(optarg);
			cgipat = optarg;
			break;

		case 'S':		/* SSL mode */
			dossl++;
			lport = 443;
			break;

		case 's':		/* Certificate */
			GETARG(optarg);
			certfile = optarg;
			break;

		case 'u':		/* Runtime UID */
			GETARG(optarg);
			uid = optarg;
			break;
		
		case 'e':		/* Error file */
			GETARG(optarg);
			errfile = optarg;
			break;

		case 'D':		/* No directory listing */
			list_dirs = 0;
			break;

		case 'N':		/* Server Name */
			GETARG(optarg);
			domain = optarg;
			break;

		case 'm':		/* Mime types file */
			GETARG(optarg);
			mimefile = optarg;
			break;

		case 'v':		/* Debig flag */
			debug++;
			break;

		default:
			usage(argv[0]);
			break;
		}

	shttpd_init();

	if (inetd) {
		/* In inetd mode, stdin is the connected socket. */
		newconnection(0, NULL);
		lsn = -1;
	} else {
		lsn = shttpd_open_port(lport);
	}
	
	/* Switch to alternate UID, it is safe now, after shttpd_open_port() */
#ifndef _WIN32
	if (uid) {
		struct passwd	*pw;

		if ((pw = getpwnam(uid)) == NULL)
			elog(1, "main: unknown user [%s]", uid);
		else if (setgid(pw->pw_gid) == -1)
			elog(1, "main: setgid(%s): %s", uid, strerror(ERRNO));
		else if (setuid(pw->pw_uid) == -1)
			elog(1, "main: setuid(%s): %s", uid, strerror(ERRNO));
		endpwent();
	}
#endif /* _WIN32 */

	(void) signal(SIGTERM, sigterm);
	(void) signal(SIGINT, sigterm);
#ifndef _WIN32
	(void) signal(SIGCHLD, sigchild);
	(void) signal(SIGPIPE, SIG_IGN);
#endif /* _WIN32 */
	
	if (inetd == 0)
		(void) printf("shttpd version %s "
		    "started on port %d\n", VERSION, lport);

	while (quit == 0)
		shttpd_poll(lsn, 1000);

	shttpd_fini();

	/* Close listening socket */
	if (lsn != -1)
		(void) close(lsn);

	if (inetd == 0)
		(void) printf("Exiting. %d concurrent connections.\n", nconns);
	
	return (EXIT_SUCCESS);
}
#endif /* !EMBEDDED */
